
signature COMPOSE = sig
    val compose : TTS.TTS -> TTS.TTS
end;
 


structure Compose : COMPOSE = struct


(* for storing architecture. Needed for -arch and possibly for translation of probes *)
structure Arch = struct

open OutPrint

datatype arch =
	 Leaf of string * (string * string) list * string list  (* process instance, with shvar renaming (for probes) and states *)
       | Node of string * arch list                             (* component, with attached instances *)

val ATREE = ref ([] : arch list);

fun push_leaf s r stl =
    ATREE := Leaf (s,r,stl) :: !ATREE;

fun push_node n a =
    let fun pop 0 l = l
	  | pop n l = let val h::t = !ATREE in ATREE := t; pop (n-1) (h::l) end
    in ATREE := Node (n, pop a []) :: !ATREE
    end;

val ANON = ref 0;
fun anon () =
    (ANON := 1 + !ANON; "_ANON_"^Int.toString (!ANON));

fun push_anon n a =
    let fun pop 0 l = l
	  | pop n l = let val h::t = !ATREE in ATREE := t; pop (n-1) (h::l) end
    in ATREE := Node (n,pop a []) :: !ATREE
    end;

(* for -arch, bottom up, as before *)
fun print_arch_bu () =
    if !ARCH then
	let fun name (Node (n,_)) = n | name (Leaf (n,_,_)) = n
	    fun prn (Node (n,sl)) =
		(List.app prn sl;
		 print n; List.app (fn s => (print " "; print (name s))) sl; print "\n")
	      | prn (Leaf (s,_,_)) =
		(print s; print "\n")
	in print "Component hierarchy:\n";
	   prn (hd (!ATREE));
	   print "\n";
	   flush ()
	end
    else
	();

(* top down *)
fun print_arch_td () =
    if !ARCH then
	let fun newline n = if n<0 then () else (print "\n"; repeat n print " ")
	    fun prn i (Node (n,sl)) = (newline i; print n; List.app (prn (i+2)) sl)
	      | prn i (Leaf (s,_,_)) = (newline i; print s)
	in print "Component hierarchy:\n";
	   prn 0 (hd (!ATREE));
	   print "\n";
	   flush ()
	end
    else
	();

end;



(* handling of temporary variables, parameter passing, and simplification:

  step 1 (per declaration):
   - preprocess to sequential transitions, removing all code after StTo and transitions without StTo
   - initialize list of temporary variables to the argument list (those passed by value)

  step 2 (parameter passing):
  in caller component:
    - vars passed as reference are part of state, others are in temporary list
    - expressions passed as parameters can ALWAYS be statically evaluated in caller context
  in callee, whether process or component:
	  if variable is "by ref"
          then replace var name by name of shared var in body
	  else initialize it to the value passed

  step 3 (simplify, after composition):
   - determine which variables in body of process are indeed state variables, then:
     * move local variables not part of state to temporary variables
     * move temp variables part of state to local variables
     * inline variables not part of state and of simple types

  step 4 (emit):
     the temporary variables are created and initialized, but outside of state structure
*)


(* doubles any "_" in name to prevent ambiguity of instantiated names: *)
fun encode id = String.translate (fn #"_" => "__" | c => Char.toString c) id;

(* returns next available integer for identifier id *)
val aliasInstance =
    let val created = ref ([] : (string * int) list)
	fun root ag = hd (String.fields (fn  #":" => true | _ => false) ag)
    in fn id => let val id = root id
		in case List.partition (fn (y,x) => y=id) (!created)  of
		       (nil,neg) => (created := (id,1) :: neg; encode id^"_1")
		     | ((_,h)::_,neg) => (created := (id,h+1) :: neg; encode id^"_"^Int.toString(h+1))
		end
    end;

(* for translating probes *)
val MAINSMAP = ref ([] : (string * string) list);




(* instance names found in ATREE *) (* #### couldn't we use the instances saved in ENV instead ? #### *)
fun instance (main,path) =
    let open Arch
	fun find [] (Node (n,sl)) = (n,[],[])
	  | find (h::t) (Node (n,sl)) = find t (List.nth (sl,h-1))
	  | find _ (Leaf (n,al,stl)) = (n,al,stl)
	val SOME (_,m) = List.find (fn (x,_) => x=main) (!MAINSMAP)
	val SOME m' = List.find (fn Node (x,_) => x=m) (!ATREE)
    in find (case String.sub(m,0) of #"_" => (* pseudo _main_X added *) 1 :: path | _ => path) m'
    end handle BYE => raise BYE | _ => failwith "renameProbe.instance";




(* uniquely renames:
     process name, transition tags, variable names (including params passed by value) and states in processes
     component name, variable names (including params passed by value) and ports in components
*)
local
fun renameStm varsMap rename s = (* renames variables and states in a statement *)
    let fun renStm stt =
	    case stt of
		StSelect a1 => StSelect(map (map renStm) a1)
	      | StSequence (a1,a2) => StSequence(renStm a1,renStm a2)
	      | StCond (e,a1,a2) => StCond(aliasvExp varsMap e,renStm a1,renStm a2)
	      | StGuard e => StGuard(aliasvExp varsMap e)
	      | StWhile(a2,a3) => StWhile(aliasvExp varsMap a2,renStm a3)
	      | StForeach(a2,a3) => StForeach(aliasvExp varsMap a2,renStm a3)
	      | StAssign (a2,a3) => StAssign (aliasvExp varsMap a2,aliasvExp varsMap a3) 
	      | StTo sl => StTo (map rename sl)
	      | StInput(st,tyl,pl) => StInput(st,tyl,map (aliasvExp varsMap) pl)
	      | StOutput(p,tyl,expl) => StOutput(p,tyl,map (aliasvExp varsMap) expl)
	      | StReturn e => StReturn (aliasvExp varsMap e)
	      | _ => stt
    in renStm s
    end;
(* renames variables, states and ports in a probe bound to a port *)
fun renameProbe NONE = NONE
  | renameProbe (SOME pb)  =
    let (* eliminates paths in probes, replaces by renamed vars/states/ports in probes *)
	fun get b =
	    case b of
		StateProbe (path,s) => let val (inst,_,stl) = instance path
					   val no = index (inst ^ "_s" ^ encode s) stl
				       in (* root in path kept for next pass *)
					   ValueProbe ((#1 path,[]),InfixExp (EQ,BoolType,VarExp(inst ^ "_vstates",NatType),IntExp(no,NatType)))
				       end
	      | ValueProbe (path,e) => let val (pr,vmap,_) = instance path
					   fun varsMap v = assoc v vmap handle _ => (pr ^ "_v" ^ encode v)
                               	       in (* root in path kept for next pass *)
 					  ValueProbe ((#1 path,[]),aliasvExpf varsMap e)
				       end
	      | EventProbe (path,p) => (* root in path kept for next pass *)
 		                       EventProbe ((#1 path,[]),#1 (instance path) ^ "_p" ^ encode p)
	      | TagProbe (path,g)   => (* root in path kept for next pass *)
 		                       TagProbe ((#1 path,[]),#1 (instance path) ^ "_" ^ encode g)
	      | EnterProbe pb => EnterProbe (get pb)
	      | LeaveProbe pb => LeaveProbe (get pb)
	      | AndProbe (pb1,pb2) => AndProbe (get pb1, get pb2)
	      | OrProbe (pb1,pb2) => OrProbe (get pb1, get pb2)
	      | ImpProbe (pb1,pb2) => ImpProbe (get pb1, get pb2)
	      | NotProbe pb => NotProbe (get pb)
    in SOME (get pb)
    end;
in
fun instantiate (ProcessDecl (b,c,d,stl,rk,k,f,h,i,init,g)) =
    (* renames states, all variables and transitions *)
    (* as of .41, also renames priority relation since can be set in precompose *)
    let val instName = aliasInstance b
	fun renames s = instName ^ "_s" ^ encode s
	fun renamev s = instName ^ "_v" ^ encode s
	fun renamet s = instName ^ "_" ^ encode s
	val varsMap = List.mapPartial (fn (ValArg v,_) => SOME (v,renamev v) | _ => NONE) d @
		      map (fn (n,_,_) => (n,renamev n)) f
	val transCounter = ref ~1;
	val tagMap = ref []
	fun renameTrans (([tag],lab,atags),from,stm,b,p) =
	    (transCounter := !transCounter + 1;
	     let val trname = instName ^ "_t" ^ Int.toString(!transCounter)
	     in tagMap := (tag,trname):: !tagMap;
		(([trname],lab,map renamet atags),map renames from,renameStm varsMap renames stm,b,p)
	     end)
    in  ProcessDecl (instName,
		     c,
		     map (fn (ValArg v,ty) => (ValArg (alias v varsMap),ty) | x => x) d,
		     map renames stl,
		     [],
		     map (fn (s,ty,v) => (alias s varsMap,ty,(* no vars in initial v *) v)) k,
		     map (fn (s,ty,v) => (alias s varsMap,ty,aliasvExp varsMap v)) f,
		     map (fn (p,pr,pb,bb) => (p,pr,renameProbe pb,bb)) h,
		     i,
		     renameStm varsMap renames init,
		     map (fn (tg,from,stm,b,prl) => (tg,from,stm,b,map (fn g => assoc g (!tagMap)) prl)) (map renameTrans g))
	handle BYE => raise BYE | _ => failwith "instantiate"
    end 
  | instantiate (ComponentDecl (b,c,d,kr,k,e,f,g,init,h)) =
    (* renames all variables and local ports. *)
    let val instName = aliasInstance b
	fun renamev s = instName ^ "_v" ^ encode s
	fun renamep s = instName ^ "_p" ^ encode s
	val varsMap = List.mapPartial (fn (ValArg v,_) => SOME (v,renamev v) | _ => NONE) d @
		      map (fn (n,_,_) => (n,renamev n)) e
	val localportMap = map (fn (p,_,_,_) => (p,renamep p)) f
	fun renameComp cp =
	    case cp of
		ParComp scpl => ParComp(map (fn (pl,c) => (map (fn p => alias p localportMap) pl,renameComp c)) scpl) 
	      | InstanceComp(i,pl,el) => InstanceComp(i,map (fn p => alias p localportMap) pl,map (aliasvExp varsMap) el)
    in ComponentDecl (instName,
		      c,
		      map (fn (ValArg v,ty) => (ValArg (alias v varsMap),ty) | x => x) d,
		      [],
		      map (fn (s,ty,v) => (alias s varsMap,ty,(* no vars in initial v *) v)) k,
		      map (fn (s,ty,v) => (alias s varsMap,ty,aliasvExp varsMap v)) e,
		      map (fn (p,ch,pb,b) => (alias p localportMap,ch,renameProbe pb,b)) f,
		      map (fn (l1,l2) => (map (fn p => alias p localportMap) l1,map (fn p => alias p localportMap) l2)) g,
		      renameStm varsMap (fn s => s) init,
		      renameComp h)
       handle BYE => raise BYE | _ => failwith "instantiate"
    end
  | instantiate d = failwith "instantiate";
end;



(* where instances are stored *)
val ENV = ref ([] : Declaration list);
fun lookupenv sst = Option.valOf (List.find (fn ProcessDecl (s,_,_,_,_,_,_,_,_,_,_) => sst = s
					      | ComponentDecl (s,_,_,_,_,_,_,_,_,_) => sst = s
					      | _ => false) (!ENV));


local
(* evaluates init statements, returns updated tmp and vars bindings, and StTo (initial state) *)
(* note: typer guarantees that no init statement reads or writes shared variables *)
fun evalInitStatement init env tmp' vars' =
    let val initial = ref []
	val vars' = ref vars'
	val tmp' = ref tmp'
	val store = ref (env @ map (fn (s,_,v) => (s,v)) (!tmp') @ map (fn (v,_,e) => (v,e)) (!vars'))
	exception Cannot
	exception Break
	fun chopList f [] accu = (rev accu,[])
	  | chopList f (l as (h::t)) accu = if f h then (rev accu,l) else chopList f t (h::accu)
	fun eval e = expVal (!store) e
	fun evalInit s =
	    case s of
		StSequence(s1,s2) => (evalInit s1; evalInit s2)
	      | StCond(c,s1,s2) => (case eval c of BoolExp b => evalInit (if b then s1 else s2) | _ => raise Cannot)
	      | StTo sl => (initial := sl; raise Break)
	      | StWhile(c,s) => (while case eval c of BoolExp b => b  | _ => raise Cannot do evalInit s)
	      | StForeach _ => evalInit (ToTTS.unrollForeach s)
	      | StAssign(RecordAccessExp(r,f,ty),e) =>
		(case eval r of
		     RecordExp (fl,rty) =>
		     let val e' = eval e
		     in evalInit (StAssign(r,RecordExp (map (fn (f',e) => (f',if f=f' then e' else e)) fl,rty)))
		     end
		   | _ => raise Cannot)
	      | StAssign(ArrayAccessExp(a,i,ty),e) =>
		(case eval i of
		     IntExp (k,_) =>
		     (case eval a of
			  ArrayExp (al,aty) =>
			  let val e' = eval e
			  in evalInit (StAssign(a,ArrayExp(List.tabulate(length al,fn i => if i=k then e' else List.nth (al,i)),aty)))
			  end
			| _ => raise Cannot)
		   | _ => raise Cannot)
	      | StAssign(VarExp(v,_),e) =>
		(case chopList (fn (v',_,_) => v=v') (!vars') [] of
		     (l1,(v,ty,_)::l2) => vars' := l1 @ (v,ty,eval e)::l2
		   | _ => (case chopList (fn (v',_,_) => v=v') (!tmp') [] of
			       (l1,(v,ty,_)::l2) => tmp' := l1 @ (v,ty,eval e)::l2
			     | _ => failwith "unfoldProgram4");
		 store := env @ map (fn (s,_,v) => (s,v)) (!tmp') @ map (fn (v,_,e) => (v,e)) (!vars'))
	      | StGuard _ => failwith "unexpected guard statement in init statement"
	      | StNull => ()
	      | _ => failwith "unexpected statement in init statement"
    in evalInit init
       handle
       Break => () 
     | Cannot => comperror "cannot statically evaluate init statement, functions cannot be called statically"
     | BYE => raise BYE
     | _ => failwith "evalInitStatement.evalInit";
       (!tmp',!vars',StTo (!initial))
    end
    handle BYE => raise BYE
	 | _ => comperror "cannot evaluate init statement, missing initialization ?"
in
(* flattens composition in main. The result is a component only with process intances *)
(* called once, before instance, in compose *)
(* evaluates initializations and pass parameters through temporary variables or state variables *)
(* note: at that point, process transition are NOT labelled (they will be labeled in makeComposition) *)
fun unfoldProgram (c as ComponentDecl (_,_,_,_,_,_,lports0,_,_,_)) =
    let (* portenv: port renaming (for processes)
	   shvenv : shvar renaming
	   env    : maps reachable vars (including shared) to values, for evaluations of init statements and parameters passed
	*)
	fun unfoldComponent (ComponentDecl (st,prs,pas,_,tmp,vars,lprs,prl,init,cp)) portenv shvenv env =
	    let	(* get values of tmp vars in env: *)
		val tmp' = map (fn (v,ty,_) => (v,ty,assoc v env)) tmp
		(* evaluate local vars in env extended with tmp values: *)
		val vars' = 
		    let val env' = env @ map (fn (s,_,v) => (s,v)) tmp'
		    in map (fn (v,ty,e) => (v,ty,expVal env' e)) vars
		    end
		(* evaluate init clause in env'', this may update env' and vars' *)
		val (tmp',vars',initial) = evalInitStatement init env tmp' vars'
		(* updated environment with both tmp and local vars: *)
		val env'' = env @ map (fn (s,_,v) => (s,v)) tmp' @ map (fn (s,_,v) => (s,v)) vars'
		(* initialize cumulated lists for composition *)
		val ltmp = ref tmp' (* tmp variable declarations *)
	        val lvars = ref vars' (* local variable declarations *)
		val lports = ref lprs (* local port declarations *)
		(* unfold composition *)
		fun unfoldComp (ParComp cpl) prio =
		    (* PrioSpec inserted at most once in a component (when prio not empty) *)
		    let val c = ParComp (map (fn (x,y) =>
						 let val xy = (map (fn x => alias x portenv) x,unfoldComp y [])
						 in case y of
							ParComp sl => Arch.push_anon (Arch.anon ()) (List.length sl)
						      | _ => ();
						    xy
						 end) cpl)
		    in if null prio then c else PrioSpec (prio,c)
		    end
		  | unfoldComp (InstanceComp (id,pl,el)) _ = 
			let (* evaluate ports and values passed: *)
			    val pvalues = map (fn x => alias x portenv) pl
			    (* renames shared variables in instance parameters *)
			    val evalues = map (fn RefExp (v,ty) => RefExp (alias v shvenv,ty) | e => expVal env'' e) el
			in case lookupenv id of
				cp as ComponentDecl _ => (* instance est un composant *)
				let val c' as ComponentDecl (n',cprs,cpas,_,_,_,_,_,_,_) = instantiate cp
				    val shvEnv = List.mapPartial (fn ((RefArg v',_),RefExp (v,_)) => SOME (v',v)
								   | _ => NONE) (ListPair.zipEq (cpas,evalues))
				    val Env = List.mapPartial (fn ((ValArg v',_),e) => SOME (v',e)
								| ((RefArg v',_),RefExp (v,ty)) => SOME (v',assoc v env'')
								  handle _ => comperror ("shared variable "^v^" must be initialized"))
							      (ListPair.zipEq (cpas,evalues)) @ env''
				    val c'' as ComponentDecl (_,pargs,_,_,ctmp,cvars,clprs,_,_,c) = unfoldComponent c' [] shvEnv Env
				in  ltmp := !ltmp @ ctmp;
				    lvars := !lvars @ cvars;
				    lports := !lports @ clprs;
				    RenSpec (ListPair.zip (map fst pargs,pl),c)
				end
			      | cp as ProcessDecl _ => (* instance est un process *)
				let val c' as ProcessDecl (n',cprs,cpas,_,_,_,_,_,_,_,_) = instantiate cp
				    val portEnv = map (fn ((id',_),id) => (id',id)) (ListPair.zipEq (cprs,pvalues))
				    val shvEnv = List.mapPartial (fn ((RefArg v',_),RefExp (v,_)) => SOME (v',v)
								   | _ => NONE) (ListPair.zipEq (cpas,evalues))
				    val Env = List.mapPartial (fn ((ValArg v',_),e) => SOME (v',e)
								| ((RefArg v',_),RefExp (v,ty)) => SOME (v',assoc v env'')
								  handle _ => comperror ("shared variable "^v^" must be initialized"))
							      (ListPair.zipEq (cpas,evalues)) @ env''
				    val ProcessDecl (id',_,_,_,_,_,_,_,_,_,_) = unfoldComponent c' portEnv shvEnv Env
				in (* vars cumulated in makeComposition, rather than here *)
				    InstanceComp (id',[],[])
				end
			end
		fun flattenPrio [] = []
		  | flattenPrio ((l1,l2)::t) =
		    flat (map (fn x => map (fn y => (alias x portenv,alias y portenv)) l2) l1) @ flattenPrio t
		val newComp = unfoldComp cp (flattenPrio prl)
		val _ = 
	            (* flat:
		    let fun reclen (_,ParComp sl) = List.foldl Int.+ 0 (map reclen sl) | reclen _ = 1
			val ParComp sl = cp
		    in Arch.push_node st (List.foldl Int.+ 0 (map reclen sl))
		    end
                    with anon: *)
		    let val ParComp sl = cp in Arch.push_node st (List.length sl) end
	    in ComponentDecl (st,prs,pas,[],!ltmp,!lvars,!lports,[],initial,newComp)
	    end
  	  | unfoldComponent (ProcessDecl (cst,_,args,csts,_,ctmp,cvars,clprs,prl,init,ctrans)) portenv shvenv env =
		let (* renames shared variables in expression *)
	            fun renameEx e = mapExp (fn VarExp (v,ty) => VarExp (alias v shvenv,ty) | e => e) e
		    (* renames shared vars and ports in statement *)
		    fun renameSt s1 = 
		      case s1 of
			StAssign(p,e) => StAssign(renameEx p,renameEx e)
		      | StSelect sl => StSelect(map (map renameSt) sl)
		      | StSequence(a1,a2) => StSequence(renameSt a1,renameSt a2)
		      | StCond (e,a1,a2) => StCond(renameEx e,renameSt a1,renameSt a2)
		      | StGuard e => StGuard(renameEx e)
		      | StWhile(a2,a3) => StWhile(renameEx a2,renameSt a3)
		      | StForeach(v,a3) => StForeach(renameEx v,renameSt a3)
		      | StSignal p => StSignal(alias p portenv)
		      | StInput(p,a3,a4) => StInput(alias p portenv,a3,(* a4 *) map renameEx a4)
		      | StOutput(p,a3,a5) => StOutput(alias p portenv,a3,map renameEx a5)
		      | _ => s1
		   (* initialize temporaries to values passed, evaluate local vars, and update transitions: *)
		   val ctmp' = map (fn (v,ty,_) => (v,ty,assoc v env)) ctmp
		   val vars' = 
		       let val env' = env @ map (fn (s,_,v) => (s,v)) ctmp'
		       in map (fn (v,ty,e) => (v,ty,expVal env' e)) cvars
		       end
		   (* evaluate init clause in env'', this may update env' and vars' *)
		   val (ctmp',vars',initial) = evalInitStatement init env ctmp' vars'
		   val p = ProcessDecl (cst,
					[],
					[],
					csts,
					[],
					ctmp',
					vars',
					clprs,
					prl,
					initial,
					map (fn (a,c,d,e,f) => (a,c,renameSt d,e,f)) ctrans)

	        in ENV := p :: !ENV;
		   Arch.push_leaf cst (List.mapPartial (fn (RefArg v,ty) => SOME (v,alias v shvenv) | _ => NONE) args) csts;
		   p
	        end
	val uc = unfoldComponent c [] [] [] handle BYE => raise BYE | _ => failwith "unfoldProgram"
    in if !ARCH then Arch.print_arch_bu () else ();
       uc
    end

  | unfoldProgram d = d;
end;




(* computes composition *)
fun makeComposition (c as ComponentDecl (st,prs,pas,rtmp,tmp,vars,lprs,pris,init,cp)) =
    let	(* concats two statements, preserving normalized property for sequences *)
	fun joinTrans (s1,StNull) = s1
	  | joinTrans (StNull,s2) = s2
	  | joinTrans (s1,s2) =
	    let fun join s =
		    case s of
			StSequence(s1,s2) => StSequence(s1,join s2)
		      | StNull => s2
		      | _ => StSequence (s,s2)
	    in join s1
	    end handle _ => failwith "joinTrans"
	fun breakTrans s =
	    let (* splits s into (s1,s',s2), where (s1;s';s2)==s and s' is the first st in s satisfying f (StNull if none) *)
		(* applied after normSeq *)
		fun chop f s =
		    let fun toSeq [] = StNull | toSeq [s] = s | toSeq (h::t) = StSequence (h,toSeq t)
			fun rchop s l1 =
			    case s of
				StSequence (s1,s2) => if f s1 then (toSeq (rev l1),s1,s2) else rchop s2 (s1::l1)
			      | s => if f s then (toSeq (rev l1),s,StNull) else (toSeq (rev (s::l1)),StNull,StNull)
		    in rchop s []
		    end
		fun break s =
		    case chop (fn StTo _ => true | _ => false) s of
			(_,StNull,_) => (* no target state *) comperror "transition with no target state" (* (StNull,StNull,StNull,[]) *)
		      | (s1,StTo tg,_) =>
			(case chop (fn StInput _ => true | StOutput _ => true | StInputList _ => true | StSignal _ => true | _ => false) s1 of
			     (ss1,StNull,_) => (* no action *) (ss1,StNull,StNull,tg)
			   | (ss1,act,ss2) => (ss1,act,ss2,tg))
	    in break s
	    end handle BYE => raise BYE | _ => failwith "breakTrans"
	fun label s = case s of StInput(p,_,_) => p | StInputList(p,_) => p | StOutput(p,_,_) => p | StSignal p => p | _ => "";
	(* computes the transitions of a parComposition  e1->p1 || .. || en->pn; in which each pi is a ProcessDecl (set of transitions, recursively)
	   step 1: for each process ei-pi, retain the transitions with labels not in ei
           step 2: for each label l in union(ei):
                     get for each process having l in e the transitions labelled by l ==> [[T1],...[Tn]] : transition list list
                     retain the synchronized product of these transitions (raise error if some list member is empty ?, meaning that that interaction is dead)
                       where the synchronised product is the set {t1 x ... x tn | ti in Ti}
	   return the union of transitions kept at steps 1 and 2, encapsulated in a process whose states and local vars are updated
	 *)
	fun composeBlock (ParComp cpl) = 
	    let	(* compose each member of cpl, recursively *)
		(* ### improve: current composeBlock breaks transition statements on start and repacks them before
	           exit (recursively). It would be more effficient to break once each transition and repack once at
		   the end of makeComposition ### *)
		val procList = map (fn (x,y) => (x,composeBlock y)) cpl
		(* expose content of processes, with their transitions broken *)
		val broken = map (fn (pl, ProcessDecl (_,_,_,_,_,_,_,_,_,_,trl)) =>
				     (pl,map (fn ((tgl,_,atg),fr,s,bn,pr) => let val br as (_,act,_,_) = breakTrans s
									   in (((tgl,label act,atg),fr,s,bn,pr),br)
									   end) trl)) procList
   		(* step 1 : find union of transitions non synchronized: *)
		(* transition are rebuilt so that unreachable statements (after "to") are removed *)
		val notsynched =
		    flat (map (fn (pl,trl) => List.mapPartial (fn ((tlb as (tag,lb,_),fr,_,bn,pr),(bef,act,aft,tg)) =>
								  if mem lb pl then NONE
								  else SOME (tlb,fr,joinTrans(bef,joinTrans(act,joinTrans(aft,StTo tg))),bn,pr)) trl) broken)
		(* step 2 : *)
		(* all labels in port lists *)
		val labels = makeset (flat (map fst procList))
		(* transitions synchronizing on label lab : *)
		fun sync lab =
		    let	(* get transitions synchronizing on lab: *)
			val trset = List.mapPartial (fn (pl,trl) => if mem lab pl
								    then SOME (case List.filter (fn (((_,lb,_),_,_,_,_),(_,act,_,_)) => lb=lab) trl of
										   [] => comperror ("no interactions on synchronized port "^lab)
										 | l => l)
								    else NONE)
						    broken
			(* statement list -> sequence statement *)
			fun mkSeq [] = StNull | mkSeq [s] = s | mkSeq (h::t) = StSequence (h,mkSeq t)
			(* builds communication statement *)
			(* returns Output if present in interaction, required to reject open input in removePorts *)
			(* for convenience, special case where reception pattern is a literal handled as an output *)
			fun isliteral pl = List.all (fn BoolExp _ => true | IntExp _ => true | _ => false) pl
			fun isAny (AnyExp _) = true | isAny _ = false
			fun mkComm (OutSignal,InSignal) =
			    case (OutSignal,InSignal) of
				(StOutput(st1,pr11,expl1),StInput(st2,pr21,expl2)) =>
			      	let val expl1c = map (fn (AnyExp l,y) => AnyExp (y::l) | (x,y) => x) (ListPair.zip(expl1,expl2))
				    val assList = map (fn (x,y) => ToTTS.convAssignMatch (StAssign(if isAny y then y else x,y))) (ListPair.zip (expl2,expl1))
				in (StOutput(st1,pr11,expl1c),List.foldr StSequence StNull assList)
				end
			      | (StOutput(st1,pr11,expl1),StInputList(st2,stl2)) =>
			      	let val expl1c = List.foldr (fn (e,l) => map (fn (AnyExp l,y) => (AnyExp (y::l)) | (x,y) => x) (ListPair.zip(expl1,l))) expl1 stl2
				    val assList = flat (map (fn stList => map (fn (x,y) => ToTTS.convAssignMatch(StAssign(if isAny y then y else x,y))) (ListPair.zip (stList,expl1))) stl2)
				in (StOutput(st1,pr11,expl1c),List.foldr StSequence StNull assList)
				end
			      | (StInput(st2,pr21,stl2),StOutput(st1,pr11,expl1)) =>
				mkComm (InSignal, OutSignal)
			      | (StInputList(st2,stl2),StOutput(st1,pr11,expl1)) =>
				mkComm (InSignal, OutSignal)
			      | (StOutput(st1,pr11,expl1),StOutput(st2,pr21,expl2)) =>
				let val (expl1c,expl2c) = ListPair.unzip (map (fn (x,y) => case (x,y) of 
											       (AnyExp l1,AnyExp l2) => let val l = AnyExp (l1@l2) in (l,l) end
											     | (AnyExp l, _) => (y,y)
											     | (_, AnyExp l) => (x,x)
											     | _ => (x,y))
									      (ListPair.zip(expl1,expl2)))
				    val assList = flat (map (fn (x,y) => 
								case (x,y) of
								    (AnyExp _, AnyExp _) => []
								  | (AnyExp l, _) => map (fn x => ToTTS.convAssignMatch(StAssign(x,y))) l
								  | (_, AnyExp l) => map (fn y => ToTTS.convAssignMatch(StAssign(y,x))) l
								  | _ => [])
							    (ListPair.zip (expl1,expl2)))
				    val condList = flat (map (fn (x,(ty,y)) => if isAny x orelse isAny y orelse x=y then [] else [InfixExp(EQ,ty,x,y)]) (ListPair.zip (expl2c,ListPair.zip(pr11,expl1c))))
				in (StOutput(st1,pr11,expl1c),List.foldr StSequence (StGuard (List.foldr (fn (x,y) => InfixExp(AND,BoolType,x,y)) (BoolExp true) condList)) assList)
				end
			      | (o1 as StInput(st1,pr11,stl1),o2 as StInput(st2,pr12,stl2)) =>
				if isliteral stl1 then mkComm (StOutput(st1,pr11,stl1),o2) else
				if isliteral stl2 then mkComm (o1,StOutput(st2,pr12,stl2)) else
				(StInputList(st1,[stl1]@[stl2]),StNull)
			      | (o1 as StInputList(st1,stl1),StInput(st2,pr21,stl2)) =>
				if isliteral stl2 then mkComm (o1,StOutput(st2,pr21,stl2)) else (StInputList(st1,stl1@[stl2]),StNull)
			      | (StInput(st2,pr21,stl2),o2 as StInputList(st1,stl1)) =>
				if isliteral stl2 then mkComm (StOutput(st2,pr21,stl2),o2) else (StInputList(st1,stl1@[stl2]),StNull)
			      | (StInputList(st2,stl2),StInputList(st1,stl1)) => (StInputList(st1,stl1@stl2),StNull)
			      | (StSignal _,StSignal _) => (OutSignal,StNull)
			      | _ => failwith "makeComposition.mkComm"
			(* synchonizes transitions *)
			fun synctuple tuple =
			    let fun unzip [((tagslabatags,from,_,bnd,prl),x)] = ((tagslabatags,from,bnd,prl),x)
				  | unzip ((((tag,lab,atags),from,_,bnd,prl),(bef,act,aft,tgt))::t) =
				    let val (((tag',_,atags'),from',bnd',prl'),(bef',act',aft',tgt')) = unzip t
					val (comm,more) = mkComm(act,act')
				    in (* composing priorities:
					  composing tr  <prl>
                                          with      tr' <prl'>
				          =>        tr__tr' <all transitions in composition involving one in prl @
				                             all transitions in composition involving one in prl'>
					*)
					(((tag' @ tag,lab,atags' @ atags), from @ from',bnd,prl @ prl'),
					 (joinTrans(bef,joinTrans(bef',more)),comm,joinTrans(aft,aft'),
					  tgt @ tgt'))
				    end
				val ((tagslab,from,bnd,prl),(bef,act,aft,tgt)) = unzip tuple
			    in (tagslab,from, joinTrans (bef,joinTrans (act, joinTrans (aft,StTo tgt))),bnd,prl)
			    end
			fun tuples [x] = map (fn z => [z]) x 
			  | tuples (h::t) = flat (map (fn tp => map (fn x => x::tp) h) (tuples t))
			  | tuples [] = []
			fun check_interferences tpl =
			    (* An interaction involves at least too transitions. Each transition can be written (before,comm,after).
			       It is checked that the before parts are non-interfering, as well as the comm parts and the after parts *)
			    let val V = ref ""
				fun indep ((_,[]),(_,[])) = true
				  | indep ((r1,w1),(r2,w2)) =
				    case List.filter (fn v => mem v r2 orelse mem v w2) w1 @ List.filter (fn v => mem v r1 orelse mem v w1) w2 of
					[] => true
				      | v::_ => (V := v; false)
			    in if !Type.SHVWRIT then
				   let fun check [] = tpl
					 | check ((rw1,rw2,rw3)::t) =
					   if List.all (fn (rw1',rw2',rw3') =>
							   indep (rw1,rw1') andalso
							   indep (rw2,rw2') andalso
							   indep (rw3,rw3')) t
					   then check t
					   else if !UNSAFE
						then (compwarning ("possibly conflicting writes on shared variable "^ !V) tpl; check t)
						else (comperror ("possibly conflicting writes on shared variable "^ !V))
				   in check (map (fn (_,(s1,s2,s3,_)) => ((readInSt s1,writtenInSt s1),(readInSt s2,writtenInSt s2),(readInSt s3,writtenInSt s3))) tpl)
				   end
			       else
				   tpl
			    end
		    in map (synctuple o check_interferences) (tuples trset)
		    end
	    in ProcessDecl (let val h::t = map (fn (_,ProcessDecl (st,_,_,_,_,_,_,_,_,_,_)) => st) procList
			    in List.foldl (fn (x,y) => x ^ "_" ^ y) h t
			    end,
			    [],
			    [],
			    rev (flat (map (fn (_,ProcessDecl (_,_,_,s,_,_,_,_,_,_,_)) => s) procList)),
			    [],
			    flat (map (fn (_,ProcessDecl (_,_,_,_,_,v,_,_,_,_,_)) => v) procList),
			    flat (map (fn (_,ProcessDecl (_,_,_,_,_,_,v,_,_,_,_)) => v) procList),
			    flat (map (fn (_,ProcessDecl (_,_,_,_,_,_,_,lp,_,_,_)) => lp) procList),
			    [],
			    StTo (rev (flat (map (fn (_,ProcessDecl (_,_,_,_,_,_,_,_,_,StTo s,_)) => s) procList))),
			    flat (map sync labels) @ notsynched)
	    end

	  | composeBlock (PrioSpec (prios,comp)) =
	    let val ProcessDecl (a,b,c,d,e,f,g,h,i,j,trl) = composeBlock comp
		(* reports priorities; does NOT build closure:
		 foreach p>q :
		 let e = transitions labelled q
		 in foreach transition labelled p:
		    add set e to its "lower" list
		 end
		 *)
		fun insertPrios [] trl = trl
		  | insertPrios ((p,q)::pl) trl =
		    insertPrios pl (case flat (List.mapPartial (fn tr as ((tags,lab,_),_,_,_,_) => if lab=q then SOME tags else NONE) trl) of
					[] => trl
				      | qtags => map (fn tr as ((tags,lab,atg),fr,s,bn,pr) =>
							 if lab = p then ((tags,lab,atg),fr,s,bn,makeset (pr @ qtags)) else tr) trl)
	    in ProcessDecl (a,b,c,d,e,f,g,h,i,j,insertPrios prios trl)
	    end

	  | composeBlock (RenSpec (renaming,comp)) =
	    let val ProcessDecl (a,b,c,d,e,f,g,h,i,j,trl) = composeBlock comp
		fun renamePorts s = 
		      case s of
			StSequence(a1,a2) => StSequence(renamePorts a1,renamePorts a2)
		      | StSignal p => StSignal(alias p renaming)
		      | StInput(p,a3,a4) => StInput(alias p renaming,a3,a4)
		      | StInputList(p,l) => StInputList(alias p renaming,l)
		      | StOutput(p,a3,a5) => StOutput(alias p renaming,a3,a5)
		      | _ => s
		fun renameTrans (tr as ((tags,lab,atg),a,b,c,d)) =
		    if alias lab renaming = lab then tr else ((tags,alias lab renaming,atg),a,renamePorts b,c,d)
	    in ProcessDecl (a,b,c,d,e,f,g,h,i,j,map renameTrans trl)
	    end

	  | composeBlock (InstanceComp (sst,_,_)) = lookupenv sst

	val ProcessDecl (tag,p,a,s,_,k,v,x,_,init,trl) = composeBlock cp handle BYE => raise BYE | _ => failwith "composition"

    in ProcessDecl (tag,p,a,s,[],k @ tmp,v @ vars,x @ lprs,[],init,trl)
    end

  | makeComposition d = d;



(* renaming properties *)

(* renaming a probe found in a property *)
(* eliminates paths in probes, replaces by renamed vars/states/ports in probes *)
(* next pass will convert enter/leave/value probes to a simple value probe on an auxiliary variable *)
fun renameProbeInProp pb =
    let 
	fun get el b =
	    case b of
		StateProbe (path,s) => (* root in path kept for next pass *)
		(* converted to a value probe if found under enter or leave *)
		(* converted here rather than in postcompose because we need the state number,
                   purposedly made available in the ATREE entry for the process instance ... *)
		let val (inst,_,stl) = instance path
		in if el then
		       let val no = index (inst ^ "_s" ^ encode s) stl
		       in ValueProbe ((#1 path,[]),InfixExp (EQ,BoolType,VarExp(inst ^ "_vstates",NatType),IntExp(no,NatType)))
		       end
		   else
		       StateProbe ((inst,[]),inst ^ "_s" ^ encode s)
		end
	      | ValueProbe (path,e) => (* root in path kept for next pass *)
		let val (pr,vmap,_) = instance path
		    fun varsMap v = assoc v vmap handle _ => (pr ^ "_v" ^ encode v)
                in ValueProbe ((#1 path,[]),aliasvExpf varsMap e)
		end
	      | EventProbe (path,p) => (* root in path kept for next pass *)
 		EventProbe ((#1 path,[]),#1 (instance path) ^ "_p" ^ encode p)
	      | TagProbe (path,g) => (* root in path kept for next pass *)
		TagProbe ((#1 path,[]),#1 (instance path) ^ "_" ^ encode g)
	      | EnterProbe pb => EnterProbe (get true pb)
	      | LeaveProbe pb => LeaveProbe (get true pb)
	      | AndProbe (pb1,pb2) => AndProbe (get el pb1, get el pb2)
	      | OrProbe (pb1,pb2) => OrProbe (get el pb1, get el pb2)
	      | ImpProbe (pb1,pb2) => ImpProbe (get el pb1, get el pb2)
	      | NotProbe pb => NotProbe (get el pb)
    in get false pb
    end;

(* renames probes in a property *)
(* time constraints adjusted in frac-emit if needed, using DIGITS *)
fun renameProperty (PropertyDecl (s,p)) =
    let fun renameProp p =
	    case p of
		DeadlockFreeProp => DeadlockFreeProp
	      | InfinitelyOftenProp pb => InfinitelyOftenProp (renameProbeInProp pb)
	      | MortalProp pb => MortalProp (renameProbeInProp pb)
	      | LiveProp pb => LiveProp (renameProbeInProp pb)
	      | PresentProp (pb,io) => PresentProp (renameProbeInProp pb,io)
	      | PresentLastingProp (pb,b) => PresentLastingProp (renameProbeInProp pb,b)
	      | PresentAfterProp (pb1,pb2,io) => PresentAfterProp (renameProbeInProp pb1,renameProbeInProp pb2,io)
	      | PresentBeforeProp (pb1,pb2,io) => PresentBeforeProp (renameProbeInProp pb1,renameProbeInProp pb2,io)
	      | PresentBetweenProp (pb1,pb2,pb3,io) => PresentBetweenProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,io) 
	      | PresentUntilProp (pb1,pb2,pb3,io) => PresentUntilProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,io)
	      | AbsentProp (pb,io) => AbsentProp (renameProbeInProp pb,io)
	      | AbsentLastingProp (pb,b) => AbsentLastingProp (renameProbeInProp pb,b)
	      | AbsentAfterProp (pb1,pb2,bo) => AbsentAfterProp (renameProbeInProp pb1,renameProbeInProp pb2,bo)
	      | AbsentBeforeProp (pb1,pb2,bo) => AbsentBeforeProp (renameProbeInProp pb1,renameProbeInProp pb2,bo)
	      | AbsentBetweenProp (pb1,pb2,pb3,io) => AbsentBetweenProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,io)
	      | AbsentUntilProp (pb1,pb2,pb3,io) => AbsentUntilProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,io)
	      | LeadstoProp (pb1,pb2,io) => LeadstoProp (renameProbeInProp pb1,renameProbeInProp pb2,io)
	      | LeadstoBeforeProp (pb1,pb2,pb3,io) => LeadstoBeforeProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,io)
	      | LeadstoAfterProp (pb1,pb2,pb3,io) => LeadstoAfterProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,io)
	      | LeadstoBetweenProp (pb1,pb2,pb3,pb4,io) => LeadstoBetweenProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,renameProbeInProp pb4,io)
	      | LeadstoUntilProp (pb1,pb2,pb3,pb4,io) => LeadstoUntilProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,renameProbeInProp pb4,io)
	      | AlwaysProp (pb,io) => AlwaysProp (renameProbeInProp pb,io)
	      | AlwaysLastingProp (pb,b) => AlwaysLastingProp (renameProbeInProp pb,b)
	      | AlwaysAfterProp (pb1,pb2,bo) => AlwaysAfterProp (renameProbeInProp pb1,renameProbeInProp pb2,bo)
	      | AlwaysBeforeProp (pb1,pb2,bo) => AlwaysBeforeProp (renameProbeInProp pb1,renameProbeInProp pb2,bo)
	      | AlwaysBetweenProp (pb1,pb2,pb3,io) => AlwaysBetweenProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,io)
	      | AlwaysUntilProp (pb1,pb2,pb3,io) => AlwaysUntilProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,io)
	      | PrecedesProp (pb1,pb2,bo) => PrecedesProp (renameProbeInProp pb1,renameProbeInProp pb2,bo)
	      | PrecedesBeforeProp (pb1,pb2,pb3,bo) => PrecedesBeforeProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,bo)
	      | PrecedesAfterProp (pb1,pb2,pb3,bo) => PrecedesAfterProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,bo)
	      | PrecedesBetweenProp (pb1,pb2,pb3,pb4,bo) => PrecedesBetweenProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,renameProbeInProp pb4,bo)
	      | PrecedesUntilProp (pb1,pb2,pb3,pb4,bo) => PrecedesUntilProp (renameProbeInProp pb1,renameProbeInProp pb2,renameProbeInProp pb3,renameProbeInProp pb4,bo)
	      | AndProp (p1,p2) => AndProp (renameProp p1,renameProp p2)
	      | OrProp (p1,p2) => OrProp (renameProp p1,renameProp p2)
	      | ImpProp (p1,p2) => ImpProp (renameProp p1,renameProp p2)
	      | NotProp p => NotProp (renameProp p)
	      | BoxProp p => BoxProp (renameProp p)
	      | DiaProp p => DiaProp (renameProp p)
	      | NextProp p => NextProp (renameProp p)
	      | UntilProp (p1,p2) => UntilProp (renameProp p1,renameProp p2)
	      | ReleaseProp (p1,p2) => ReleaseProp (renameProp p1,renameProp p2)
	      | ProbeProp p => ProbeProp (renameProbeInProp p)
	      | VarProp s => VarProp s
    in PropertyDecl (s,renameProp p)
    end
  | renameProperty d = d;




(* since dummy main component added in frac-precompose if needed, top of section is always a component;
   makeComposition always returns a process *)
fun compose (Program sl) =
    let val c = List.length sl
	fun compSection ("#properties",dl,m) =
	    ("#properties",List.map renameProperty dl,m)
	  | compSection (_,dl,Main (main,low)) =
	    let val funs = List.filter (fn FunctionDecl _ => true | ExternDecl _ => true | _ => false) dl
		val r = lookupenv main
		val r' as ComponentDecl (main',_,_,_,_,_,_,_,_,_) = if c <= 1 then r else instantiate r
	    in MAINSMAP := (main,main') :: !MAINSMAP; (* for translating probe roots *)
	       (* update inter-section priority relation for post-compose *)
	       INTERPRIO := List.map (fn s => (main,s)) low @ !INTERPRIO;
	       (main, funs, makeComposition (unfoldProgram r'))
	    end
    in ENV := flat (map (fn (_,x,_) => x) sl);
       Program (map compSection sl)
    end
    handle BYE => raise BYE | _ => failwith "compose";


end; (* Compose *)

