
(* prints error message on stderr and aborts *)
fun emit_error m = let open ErrPrint in print "emit error: "; print m; print "\n"; flush (); raise BYE end;



(*
applied before emit, normalizes the types of state variables,
introduicing the necessary type declarations;
- interval and nat types replaced by int;
- unions, records and queues are named (but not arrays);
*)

signature NAME = sig
  val nameTypes : TTS -> TTS
end

structure Name :> NAME = struct

(* utilities *)

(* lexicographic list comparisons *)
fun rcompare _ [] (_ :: _) = LESS
  | rcompare _ (_ :: _) [] = GREATER
  | rcompare cmp (h1 :: t1) (h2 :: t2) = (case cmp (h1,h2) of EQUAL => rcompare cmp t1 t2 | b => b)
  | rcompare _ [] [] = EQUAL;

(* used to creates new names for types and channels *)
val newName =
    let val no = ref 0
    in fn prefix =>
           let val n = prefix ^ Int.toString (!no)
           in no := 1 + !no;
              n
           end
    end;

(* holds processed declarations *)
val DECLS = ref ([] : (string * Type) list);


(* type keyed avl of strings *)
structure typeKey : KEY = struct
type key = Type;
fun level ty =  (* identifies all numeric types *)
 case ty of
	  BoolType => 1
	| IntType => 2
	| CharType => 2
	| UnsignedCharType => 2
	| ShortType => 2
	| UnsignedShortType => 2
	| NatType => 2
	| NamedType _ => 3
	| EnumType _ => 4
	| UnionType _ => 5
	| RecordType _ => 6
	| ArrayType _ => 7
	| QueueType _ => 8
        | _ => (* unexpected *) 20;

fun compare (x,y) = case Int.compare (level x,level y) of
	  EQUAL => (case (x,y) of
		(NamedType s1, NamedType s2) => String.compare (s1,s2) 
	      | (EnumType cl1, EnumType cl2) =>	rcompare (fn (c1,c2) => String.compare (c1,c2)) cl1 cl2
	      | (UnionType ctyl1, UnionType ctyl2) =>
		rcompare (fn ((c1,ty1),(c2,ty2)) =>
			     case String.compare (c1,c2) of
				 EQUAL => (case (ty1,ty2) of
					       (NONE,NONE) => EQUAL
					     | (NONE,_) => LESS
					     | (_,NONE) => GREATER
					     | (SOME ty1,SOME ty2) => compare(ty1,ty2))
			       | b => b)
			 ctyl1
			 ctyl2
	      | (RecordType ftyl1, RecordType ftyl2) =>
		rcompare (fn ((f1,ty1),(f2,ty2)) => case String.compare (f1,f2) of EQUAL => compare(ty1,ty2) | b => b) ftyl1 ftyl2
	      | (ArrayType (sz1,ty1), ArrayType (sz2,ty2)) =>
		(case Int.compare (sz1,sz2) of EQUAL => compare (ty1,ty2) | b => b)
	      | (QueueType (sz1,ty1), QueueType (sz2,ty2)) =>
		(case Int.compare (sz1,sz2) of EQUAL => compare (ty1,ty2) | b => b)
	      | _ => EQUAL)
	| v => v;

fun x < y = case compare (x,y) of LESS => true | _ => false;
fun x > y = case compare (x,y) of GREATER => true | _ => false;
end;
structure tyAvlTree = avlTreef (typeKey);

(* names avl : type -> name, if any *)
val TYNAMES = tyAvlTree.newTree() : String.string tyAvlTree.tree ref;
fun insertType ty n = TYNAMES := tyAvlTree.insert (ty,n,!TYNAMES);
fun lookupType ty = tyAvlTree.getData (ty,!TYNAMES);

(* type ty was found equivalent to n (modulo arithmetic types), updates definition of n if not equal *)
fun adjustTypeDecl n ty =
    let infix lub
	fun ty1 lub ty2 = (* ty2 can be named *)
	    case (ty1,ty2) of
	      (* one is signed *)
		(IntType,_) => IntType
	      | (_,IntType) => IntType
	      | (ShortType,_) => ShortType
	      | (_,ShortType) => ShortType
	      | (CharType,_) => CharType
	      | (_,CharType) => CharType
	      (* none is signed *)
	      | (NatType,_) => NatType
	      | (_,NatType) => NatType
	      | (UnsignedShortType,_) => UnsignedShortType
	      | (_,UnsignedShortType) => UnsignedShortType
	      | (UnsignedCharType,_) => UnsignedCharType
	      | (_,UnsignedCharType) => UnsignedCharType
	      (* structures *)
	      | (RecordType flty1, RecordType flty2) => RecordType (map (fn ((f,ty1),(_,ty2)) => (f,ty1 lub ty2)) (ListPair.zip (flty1,flty2)))
	      | (ArrayType (sz,ty1), ArrayType (_,ty2)) => ArrayType(sz,ty1 lub ty2)
	      | (QueueType (sz,ty1), QueueType (_,ty2)) => QueueType(sz,ty1 lub ty2)
	      | _ => ty2
	val ty' = assoc n (!DECLS)
	fun upd [] = [] | upd ((nty as (n',_))::t) = if n=n' then (n',ty lub ty')::t else nty::upd t
    in DECLS := upd (!DECLS)
    end;
    
(* creates a name for a type and its constituents, making sure names are new *)
fun nameType pref ty =
    (* if ty is named, return its name, otherwise generate a new name and stores it *)
    case lookupType ty of 
	SOME n => (adjustTypeDecl n ty; NamedType n)
      | _ => let val n = newName pref
	     in DECLS := (n,ty) :: !DECLS;
		insertType ty n;
		NamedType n
	     end;

(* returns alias for ty, unless is a constant type *)
(* constants have been computed, and names expanded *)
(* first version specializes representations of intervals (to save space) and keeps NatType.
   This interferes with subtyping.
   optimal solution would be to add casts in the code generated when a value of a named type is promoted to its supertype,
   suboptimal solution here is to keep only the lubs of named types in the < relation (nameType handles that) *)
fun aliasType ty =
    case ty of
        IntervalType (a,b) =>
	((* if a >= 0
	 then if b < 256 then UnsignedCharType else
	      if b < 65536 then UnsignedShortType else
	      NatType
	 else *) if a >= ~127 andalso b < 128 then CharType else
	      if a >= ~32768 andalso b < 32768 then ShortType else
	      IntType)
      | NatType => IntType
      | EnumType cl => nameType "e" ty
      | UnionType ctyl => nameType "u" (UnionType (map (fn (f,SOME ty) => (f,SOME (aliasType ty)) | co => co) ctyl))
      | RecordType ftyl => nameType "r" (RecordType (map (fn (f,ty) => (f,aliasType ty)) ftyl))
      | ArrayType (sz,ety) => nameType "a" (ArrayType (sz,aliasType ety))
      | QueueType (sz,ety) => nameType "q" (QueueType (sz,aliasType ety))
      | _ => ty;

(* names types in embedded StBlocks *)
fun nameInExp e = (* note: only types occurring in declarations are named (except in special cases ...) *)
    case e of
 	InfixExp(EQ,ty,e1,e2) => InfixExp(EQ,aliasType ty,nameInExp e1,nameInExp e2)
      | InfixExp(NE,ty,e1,e2) => InfixExp(NE,aliasType ty,nameInExp e1,nameInExp e2)
      | InfixExp(f,ty,e1,e2) => InfixExp(f,ty,nameInExp e1,nameInExp e2)
      | BinopExp(f,ty,e1,e2) => BinopExp(f,ty,nameInExp e1,nameInExp e2)
      | UnopExp(f,ty,e) => UnopExp(f,ty,nameInExp e)
      | RecordExp(fl,ty) => RecordExp(map (fn (f,e) => (f,nameInExp e)) fl,ty)
      | ArrayExp(el,ty) => ArrayExp(map nameInExp el,ty)
      | QueueExp(el,ty) => QueueExp(map nameInExp el,ty)
      | ArrayAccessExp(e1,e2,ty) => ArrayAccessExp(nameInExp e1,nameInExp e2,ty)
      | RecordAccessExp(e,f,ty) => RecordAccessExp(nameInExp e,f,ty)
      | CondExp(c,e1,e2) => CondExp(nameInExp c,nameInExp e1,nameInExp e2)
      | ConstrExp(c,e,ty) => ConstrExp(c,nameInExp e,ty)
      | AppExp(c,el,tyl,ty) => AppExp(c,map nameInExp el,tyl,ty)
      | CallExp(c,el,tyl,ty) => CallExp(c,map nameInExp el,tyl,ty)
      | DestrExp (c,e) => DestrExp (c,nameInExp e)
      | MatchesExp(c,e) => MatchesExp(c,nameInExp e)
      | LetExp(s,e) => LetExp(nameInSt s,nameInExp e)
      | _ => e
and nameInSt s =
    case s of
	StGuard e => StGuard (nameInExp e)
      | StBlock(dl,s) => StBlock (map (fn (s,ty) => (s,aliasType ty)) dl, nameInSt s)
      | StSequence(s1,s2) => StSequence(nameInSt s1, nameInSt s2)
      | StWhile(e,s) => StWhile(nameInExp e, nameInSt s)
      | StForeach(v,s) => StForeach(v, nameInSt s)  (* note: type NOT named in v *)
      | StCond(e,s1,s2) => StCond(nameInExp e, nameInSt s1, nameInSt s2)
      | StAssign(p,e) => StAssign(nameInExp p,nameInExp e)
      | StReturn e => StReturn(nameInExp e)
      | _ => s;

(* names types in declarations and in constants (so that enum/union declarations are not lost!) *)
(* added type declarations prefix first section *)
fun nameTypes (Program sl) =
    let fun nameDecl d =
	    case d of
		ProcessDecl (b,c,d,e,rkl,kl,vl,f,g,init,h) =>
		let val rkl' = map (fn (s,ty,e) => (s,aliasType ty,nameInExp e)) rkl
		    val kl' = map (fn (s,ty,e) => (s,aliasType ty,nameInExp e)) kl
		    val vl' = map (fn (s,ty,e) => (s,aliasType ty,nameInExp e)) vl
		    val h' = List.map (fn (tg,fr,s,b,p) => (tg,fr,nameInSt s,b,p)) h
		in ProcessDecl (b,c,d,e,rkl',kl',vl',f,g,init,h')
		end
	      | FunctionDecl (f,args,ty,vars,body) =>
		FunctionDecl (f,
			      map (fn (v,ty) => (v,aliasType ty)) args,
			      aliasType ty,
			      map (fn (s,ty,e) => (s,aliasType ty,nameInExp e)) vars,
			      nameInSt body)
	      | ExternDecl (f,argtypes,ty) => ExternDecl (f,map aliasType argtypes,aliasType ty)
	      | d => d
	fun nameSection (s,dl,m) =
	    (s, map nameDecl dl, nameDecl m)
	val ((s1,dl1,m1)::sect) = map nameSection sl
    in Program ((s1,map TypeDecl (rev (!DECLS)) @ dl1,m1)::sect)
    end;

end; (* structure Name *)






(* prints the abstract tts by Prn *)
functor TtsEmit (Prn : PRN) = struct

open Prn;

fun newline n = if n<0 then () else (print "\n"; repeat n print " ");

fun print_Fiacre (Program sl) =
	List.app (fn (_,dl,m) => (List.app print_Decl dl; print_Decl m)) sl
and print_Decl d =
        case d of
	    TypeDecl d => print_TypeDecl 0 d
       (* | ChannelDecl d => print_ChannelDecl 0 d *)
	  | FunctionDecl d => print_Function 0 d
	  | ExternDecl d => print_Extern 0 d
	  | ProcessDecl d => print_Process 0 d
	  | ComponentDecl d => print_Component 0 d
	  | Main (m,low) => (newline 0; print m;
			     case low of [] => () | _ => (print " > "; print_list (""," | ",")") print low))
	  | PropertyDecl d => print_Property 0 d
	  | Assert "" => ()
	  | Assert p => (newline 0; print "assert "; print p)
and print_TypeDecl c (a,b) =
	(newline c; print "type "; print a; print " is "; print_Type b; newline 0)
(*
and print_ChannelDecl c (l,a,b) =
	(newline c; print "channel "; print a; print " is "; print_Channel b; newline 0)
*)
and print_Type ty = (case ty of
	  BoolType => print "bool"
	| NatType => print "nat"
	| IntType => print "int"
	| CharType => print "char"
	| UnsignedCharType => print "unsigned char"
	| ShortType => print "short"
	| UnsignedShortType => print "unsigned short"
	| NamedType a => print a
	| IntervalType(a,b) => (printi a; print ".."; printi b)
	| EnumType a => print_list ("enum "," | "," end") print a  (* no enums shown *)
	| UnionType a => print_list ("union "," | "," end") (fn (xl,NONE) => print xl | (xl,SOME y) => (print xl; print " of "; print_Type y)) a
	| RecordType a 	=> print_list ("record ",", "," end") (fn (xl,y) => (print xl; print ":"; print_Type y)) a
	| ArrayType (e,b) => (print "array "; printi e; print " of "; print_Type b)
	| QueueType (e,b) => (print "queue "; printi e; print " of "; print_Type b))
and print_Channel ch =
        print_list (""," # ","") print_Type ch
and print_Profile tyl =
        case tyl of
	    [] => print "sync"
	  | _  => print_Channel tyl
and print_Function i (f,args,ty,vars,body) =
	(newline i; print "function "; print f;
	 print_list (" (",", ",")") (fn (v,ty) => (print v; print ": "; print_Type ty)) args;
	 print " : "; print_Type ty;
	 print " is ";
	 if null vars then () else
	 (newline (i+4); print_list ("var ",", ","") (fn (s,ty,v) => (print s; print ": "; print_Type ty; print ":= "; print_Exp v)) vars);
	 newline (i+4); print "begin";
	 print_Statement (i+6) body;
	 newline (i+4); print "end";
	 newline 0)
and print_Extern i (f,argtypes,ty) =
	(newline 0; print "extern "; print ("\""^f^"\"");
	 print_list (" (",", ",")") print_Type argtypes;
	 print " : "; print_Type ty;
	 newline 0)
and print_Process i (a,b,d,sd,rk,k,f,h,g,init,j) =
	(newline 0; print "process "; print a;
	 case b of [] => () | _ => (newline 0; newline (i+4); print_list ("[",", ","]") print_PortDecl b);
	 case d of [] => () | _ => (newline 0; newline (i+4); print_list ("{",", ","}") print_VarParam d);
	 newline (i+4); print "is";
	 newline 0; newline (i+4); print_StatesDecl sd;
	 case rk of [] => () | _ => (newline 0; newline (i+4); print_list ("const ",", ","") print_VarDecl rk);
	 case k of [] => () | _ => (newline 0; newline (i+4); print_list ("tmp ",", ","") print_VarDecl k);
	 case f of [] => () | _ => (newline 0; newline (i+4); print_list ("var ",", ","") print_VarDecl f);
	 case h of [] => ()
		 | h::t => (newline 0; newline (i+4); print "// port "; print_locPortDecl h;
			    List.app (fn p => (print ","; newline (i+4); print "//      "; print_locPortDecl p)) t);
	 case g of [] => () | _ => (newline 0; newline (i+4); print_priorities g);
	 case init of StNull => () | StTo [] => () | _ => (newline 0; newline (i+4); print "init "; print_Statement (i+4) init);
	 app (print_Transition (i+4)) j;
	 newline 0)
and print_Component i (a,b,d,rk,k,e,f,g,init,h) =
	(newline 0; print "component "; print a;
	 case b of [] => () | _ => (newline (i+4); print_list ("[",", ","]") print_PortDecl b);
	 case d of [] => () | _ => (newline (i+4); print_list ("{",", ","}") print_VarParam d);
	 newline (i+4); print "is";
	 case rk of [] => () | _ => (newline (i+4); print_list ("const ",", ","") print_VarDecl rk);
	 case k of [] => () | _ => (newline (i+4); print_list ("tmp ",", ","") print_VarDecl k);
	 case e of [] => () | _ => (newline (i+4); print_list ("var ",", ","") print_VarDecl e);
	 case f of [] => ()
		 | h::t => (newline (i+4); print "port "; print_locPortDecl h;
			    List.app (fn p => (print ","; newline (i+9); print_locPortDecl p)) t);
	 case g of [] => () | _ => (newline 0; newline (i+4); print_priorities g);
	 case init of StNull => () | StTo [] => () | _ => (newline 0; newline (i+4); print "init "; print_Statement (i+4) init);
	 print_Composition (i+4) h false;
	 newline 0)
and print_priorities [] = ()
  | print_priorities l = print_list ("priority ",", ","")
				    (fn (l1,l2) => (print_list (""," ","") print l1; print " > "; print_list (""," ","") print l2)) 
				    l
and print_locPortDecl (name,ch,pb,(left,right)) =
        (print name; print ": "; print_Profile ch; print_TimeConstraint (left,right);
	 case pb of SOME p => (print " is "; print_Probe p) | _ => ())
and print_Composition i c x = (case c of (* x true if starts at beginning of line *)
	  ParComp cl  => (if x then () else newline i; print "par";
			  let val (pl,h)::t = cl
			      val no = ref false
			  in newline (i+3);
			     case pl of [] => no:=true | _ => (print_list ("",",","") print pl; print " -> ");
			     print_Composition (i+3) h (!no);
			     List.app (fn (pl,c) =>
					  (newline i; print "|| ";
					   case pl of [] => no:=true | _ => (print_list ("",",","") print pl; print " -> ");
					   print_Composition (i+3) c (!no))) t
			  end;
			  newline i; print "end")
	| InstanceComp c => print_InstanceComp i c
	| PrioSpec (prl,c) => (if x then () else newline i;
			       print "prio "; print_list ("",", ","") (fn (x,y) => (print x; print " > "; print y)) prl;
			       print_Composition i c false)
	| RenSpec (rnl,c) => (if x then () else newline i;
			      print "renaming "; print_list ("",", ","") (fn (x,y) => (print x; print " -> "; print y)) rnl;
			      print_Composition i c false))
and print_InstanceComp i (id,pl,al) =
	(print id;
	 if null pl then () else print_list (" [",",","]") print pl;
	 if null al then () else print_list (" (",",",")") print_Exp al)
and print_PortDecl (name,ch) =
	(print name; print ":"; print_Profile ch)
and print_arg (ValArg s) = print s | print_arg (RefArg s) = print s
and print_VarParam (name,ty) =
	(print_arg name; print ":"; print_Type ty)
and print_VarDecl (a,b,c) =
	(print a; print ": "; print_Type b; print " := "; print_Exp c)
and print_StatesDecl a =
	(print_list ("states ",", ","") print a)
and print_TimeConstraint (left,right) =
        if case (left,right) of (Closed d,Infinite) => d <= 0.0 | _ => false then () else
	(print " in "; print_Interval (left,right))
and print_Interval (left,right) =
    let fun printb b = print (Int.toString (Real.floor (b * (pow (10.0,!DIGITS)))))
    in case left of Open b => (print "]"; printb b)
		  | Closed b => (print "["; printb b);
       print ",";
       case right of Open b => (printb b; print "[")
		   | Closed b => (printb b; print "]")
		   | Infinite => print "...["
    end
and print_Transition i ((tags,lab,atags),b,c,bnd,prl) =
	(newline 0; newline i; print "Trans::";
	 print (String.concatWith "__" tags);
	 case lab of "" => () | _ => print (": " ^ lab);
	 case atags of [] => () | _ => (print " & "; print (String.concatWith " " atags));
	 newline (i+2); print "from "; print_list (""," ","") print b; print " "; print_Statement (i+3) c;
	 newline (i+2); print "in "; print_Interval bnd;
	 case prl of [] => () | _ => (newline (i+2); print "> "; print_list (""," ","") print prl))
and print_Statement e st = case st of
	  StNull => (newline e; print "null")
	| StGuard a => (newline e; print "on "; print_Exp a)
	| StCond(a,b,c) =>
		(newline e; print "if "; print_Exp a; print " then "; print_Statement (e+3) b;
		 newline e; print " else ";  print_Statement (e+3) c;
		 newline e; print " end /* if */")
	| StSelect (a::t) =>
	        let fun print_alternant e (a::t) =
		  (print_Statement (e+3) a;
		   List.app (fn y => (newline e; print "[]"; print_Statement (e+3) y)) t);
	        in newline e; print "select";
		   print_alternant e a;
		   List.app (fn y => (newline e; print "unless"; print_alternant e y)) t;
		   newline e; print "end /* select */"
                 end
	| StSequence(a,b) => (print_Statement e a; print "; "; print_Statement e b)
	| StSignal a => (newline e; print a)
	| StInput (a,r,b) =>
		(newline e; print a; print_pdetails r; print "? "; print_list ("",",","") print_Patt b)
	| StInputList (a,b) =>
		(print a; print_list ("[","-","]") (fn y => print_list ("",",","") print_Exp y) b)
  	| StOutput (a,r,b) =>
		(newline e; print a; print_pdetails r; print "! "; print_list ("",", ","") print_Exp b)
	| StAssign (a,b) =>
		(newline e; print_Patt a; print " := "; print_Exp b)
	| StWhile(a,b) => (newline e; print "while "; print_Exp a; print " do "; 
			     print_Statement (e+3) b; 
			     newline e; print " end /* while */")
	| StForeach(a,b) => (newline e; print "foreach "; print_Exp a; print " do "; 
			     print_Statement (e+3) b; 
			     newline e; print " end /* foreach */")
	| StTo a => (newline e; print "to "; print_list (""," ","") print a)  (* ### loop deduced from src list ### *)
	| StBlock (dl,s) => (newline e; print "block {"; 
			     map (fn (v,ty) => (newline (e+2); print v; print " : "; print_Type ty; print "; ")) dl;
			     print_Statement (e+2) s;
			     newline e; print "} /* block */")
	| StReturn exp => (newline e; print "return "; print_Exp exp)
	| StFailwith s => (newline e; print "failwith "; print ("\""^s^"\""))
	| _ => failwith "unexpected statement in print_Statement"
and print_Patt e =
        let val d = !DETAILS in DETAILS := false; print_Exp e; DETAILS := d end
and print_Exp e = case e of
	  VarExp (s,r)          => (print s; print_details r)
	| EnumExp (s,r)         => (print s; print_details r)
	| ConExp (s,r)          => (print s; print_details r)
	| ArrayAccessExp(a,b,_) => (print_Exp a; print "["; print_Exp b; print "]")
 	| RecordAccessExp(a,b,_) => (print_Exp a; print "."; print b)
	| IntExp (n,r)          => (printi n; print_details r)
	| BoolExp b             => print (Bool.toString b)
	| InfixExp(x,r,y,z)     => (print "("; print_Exp y; print " "; print_Infix x; print_details r; print " "; print_Exp z; print ")")
	| BinopExp(x,r,y,z)	=> (print_Binop x; print_details r; print "("; print_Exp y; print ","; print_Exp z; print ")")
	| UnopExp(a,r,b)        => (print_Unop a; print_details r; print "("; print_Exp b; print ")")
	| ConstrExp(c,e2,r)     => (print c; print "("; print_Exp e2; print_details r; print ")")
	| AppExp(f,el,_,r)      => (print f; print_list (" (",",",")") print_Exp el; print_details r)
	| CallExp(f,el,_,r)     => (print ("\""^f^"\""); print_list (" (",",",")") print_Exp el; print_details r)
	| RecordExp (fl,r)      => let val b = !DETAILS
				   in DETAILS := false; print_list ("{",",","}") (fn (f,l) => (print f; print "="; print_Exp l)) fl;
				      DETAILS := b; print_details r
				   end
	| ArrayExp (el,r)       => let val b = !DETAILS
				   in DETAILS := false; print_list ("[",",","]") print_Exp el; DETAILS := b; print_details r
			           end
	| QueueExp (el,r)   	=> (print_list ("{|",",","|}") print_Exp el; print_details r)
	| CondExp (c,e1,e2)     => (print "("; print_Exp c; print ") ? ("; print_Exp e1; print ") : ("; print_Exp e2; print ")")
	| RefExp (v,r)          => (print "&"; print v; print_details r)
        | LetExp(s,e)           => (print "({"; print_Statement (~1000) s; print "; "; print_Exp e; print "})")
	| DestrExp (c,e)        => (print "Destr ("; print_Exp e; print ")")
	| MatchesExp (c,e)      => (print "Matches ("; print c; print ","; print_Exp e; print ")")
	| AnyExp []             => print "any"
	| AnyExp _              => failwith "unexpected 'any' pattern in send"
(* 	| _                     => failwith "unexpected exp in print_Exp" *)
and print_details ty =
	if !DETAILS then (print "::"; print_Type ty) else ()
and print_rdetails tyl =
	if !DETAILS then (print "::"; print_list (""," * ","") print_Type tyl) else ()
and print_pdetails ty =
	if !DETAILS then (print "::"; print_Channel ty) else ()
and print_Infix ope = print (case ope of
	  AND => "and"	| OR  => "or"	| LE  => "<="	| LT  => "<"	| GT  => ">"	| GE  => ">="
	| EQ  => "="	| NE  => "<>"	| ADD => "+"	| SUB => "-"	| MOD => "%"	| MUL => "*"	| DIV => "/")
and print_Binop ENQUEUE = print "enqueue" | print_Binop APPEND = print "append"
and print_Unop ope = print (case ope of
	  COERCE => "$" | NOT => "not" | FULL => "full" | EMPTY => "empty" | DEQUEUE => "dequeue"
	| PLUS => "+" | MINUS => "-" | FIRST => "first" | LENGTH => "length")
and print_Probe pb = (case pb of
          StateProbe (_,s) => (print "state "; print s)
        | ValueProbe (_,e) => (print "value "; print_Exp e)
        | EventProbe (_,r) => (print "event "; print r)
        | TagProbe (_,r)   => (print "tag "; print r)
        | EnterProbe r => (print "enter ("; print_Probe r; print ")")
        | LeaveProbe r => (print "leave ("; print_Probe r; print ")")
	| NotProbe r => (print "not ("; print_Probe r; print ")")
	| AndProbe (r1,r2) => (print_Probe r1; print " and "; print_Probe r2)
	| OrProbe (r1,r2) => (print_Probe r1; print " or "; print_Probe r2)
	| ImpProbe (r1,r2) => (print_Probe r1; print " => "; print_Probe r2))
and print_Bound bnd =
    case bnd of
	Closed b => print (Int.toString (Real.floor (b * (pow (10.0,!DIGITS)))))
      | Open b => print (Int.toString (Real.floor (b * (pow (10.0,!DIGITS)))))
      | _ => print "..."
and print_lasting (b as Closed _) = (print " lasting "; print_Bound b)
  | print_lasting _ = print "?"
and print_duration (SOME (b as Closed _)) = (print " for "; print_Bound b)
  | print_duration (SOME _) = (print " for "; print "?")
  | print_duration _ = ()
and print_within (SOME i) = (print " within "; print_Interval i)
  | print_within _ = ()
and print_Prop p = (case p of
         DeadlockFreeProp => print "deadlockfree"
       | InfinitelyOftenProp p => (print "infinitelyoften "; print_Probe p)
       | MortalProp p => (print "mortal "; print_Probe p)
       | LiveProp p => (print "live "; print_Probe p)
       | PresentProp (p,i) => (print "present "; print_Probe p; print_within i)
       | PresentLastingProp (p,d) => (print "present "; print_Probe p; print_lasting d)
       | PresentAfterProp (p1,p2,i) => (print "present "; print_Probe p1; print " after "; print_Probe p2; print_within i)
       | PresentBeforeProp (p1,p2,i) => (print "present "; print_Probe p1; print " before "; print_Probe p2; print_within i)
       | PresentBetweenProp (p1,p2,p3,i) => (print "present "; print_Probe p1; print " between "; print_Probe p2; print " and "; print_Probe p3; print_within i)
       | PresentUntilProp (p1,p2,p3,i) => (print "present "; print_Probe p1; print " after "; print_Probe p2; print " until "; print_Probe p3; print_within i)
       | AbsentProp (p,i) => (print "absent "; print_Probe p; print_within i)
       | AbsentLastingProp (p,d) => (print "absent "; print_Probe p; print_lasting d)
       | AbsentAfterProp (p1,p2,d) => (print "absent "; print_Probe p1; print " after "; print_Probe p2; print_duration d)
       | AbsentBeforeProp (p1,p2,d) => (print "absent "; print_Probe p1; print " before "; print_Probe p2; print_duration d)
       | AbsentBetweenProp (p1,p2,p3,i) => (print "absent "; print_Probe p1; print " between "; print_Probe p2; print " and "; print_Probe p3; print_within i)
       | AbsentUntilProp (p1,p2,p3,i) => (print "absent "; print_Probe p1; print " after "; print_Probe p2; print " until "; print_Probe p3; print_within i)
       | LeadstoProp (p1,p2,i) => (print_Probe p1; print " leadsto "; print_Probe p2; print_within i)
       | LeadstoBeforeProp (p1,p2,p3,i) => (print_Probe p1; print " leadsto "; print_Probe p2; print " before "; print_Probe p3; print_within i)
       | LeadstoAfterProp (p1,p2,p3,i) =>  (print_Probe p1; print " leadsto "; print_Probe p2; print " after "; print_Probe p3; print_within i)
       | LeadstoBetweenProp (p1,p2,p3,p4,i) => (print_Probe p1; print " leadsto "; print_Probe p2; print " between "; print_Probe p3; print " and "; print_Probe p4; print_within i)
       | LeadstoUntilProp (p1,p2,p3,p4,i) => (print_Probe p1; print " leadsto "; print_Probe p2; print " after "; print_Probe p3; print " until "; print_Probe p4; print_within i)
       | AlwaysProp (p,i) => (print "always "; print_Probe p; print_within i)
       | AlwaysLastingProp (p,b) => (print "always "; print_Probe p; print_lasting b)
       | AlwaysAfterProp (p1,p2,b) => (print "always "; print_Probe p1; print " after "; print_Probe p2; print_duration b)
       | AlwaysBeforeProp (p1,p2,b) => (print "always "; print_Probe p1; print " before "; print_Probe p2; print_duration b)
       | AlwaysBetweenProp (p1,p2,p3,i) => (print "always "; print_Probe p1; print " between "; print_Probe p2; print " and "; print_Probe p3; print_within i)
       | AlwaysUntilProp (p1,p2,p3,i) => (print "always "; print_Probe p1; print " after "; print_Probe p2; print " until "; print_Probe p3; print_within i)
       | PrecedesProp (p1,p2,b) => (print_Probe p1; print " precedes "; print_Probe p2; print_duration b)
       | PrecedesBeforeProp (p1,p2,p3,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " before "; print_Probe p3; print_duration b)
       | PrecedesAfterProp (p1,p2,p3,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " after "; print_Probe p3; print_duration b)
       | PrecedesBetweenProp (p1,p2,p3,p4,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " between "; print_Probe p3; print " and "; print_Probe p4; print_duration b)
       | PrecedesUntilProp (p1,p2,p3,p4,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " after "; print_Probe p3; print " until "; print_Probe p4; print_duration b)
       | AndProp (p1,p2) => (print_Prop p1; print " and "; print_Prop p2)
       | OrProp (p1,p2) => (print_Prop p1; print " or "; print_Prop p2)
       | ImpProp (p1,p2) => (print_Prop p1; print " => "; print_Prop p2)
       | NotProp p => (print "not "; print_Prop p)
       | BoxProp p => (print "[] "; print_Prop p)
       | DiaProp p => (print "<> "; print_Prop p)
       | NextProp p => (print "() "; print_Prop p)
       | UntilProp (p1,p2) => (print_Prop p1; print " until "; print_Prop p2)
       | ReleaseProp (p1,p2) => (print_Prop p1; print " release "; print_Prop p2)
       | ProbeProp p => (print "["; print_Probe p; print "]")
       | VarProp n => print n)
and print_Property _ (s,p) =
    (newline 0; print "property "; print s; print " is "; print_Prop p);

fun emitTts s0 = (print_Fiacre (Name.nameTypes s0); newline 0; flush ()) handle BYE => raise BYE | _ => failwith "emit tts";

end; (* functor TtsEmit *)

structure OutTtsEmit = TtsEmit(OutPrint);
structure ErrTtsEmit = TtsEmit(ErrPrint);




(* prints the TTS Net by Prn *)
structure TtsNetEmit = struct

open OutPrint;

fun newline n = if n<0 then () else (print "\n"; repeat n print " ");

fun print_Fiacre (Program ((_,_,ProcessDecl d)::_)) =
        print_Process 0 d
					       
and print_Process i (a,_,_,b,_,_,_,_,_,StTo e,c) =
	(newline i; print "net process_"; print a;
         newline i; newline i; print_StatesDecl (b,e);
	 app (print_Transition i) c;
	 app (print_prior i) c;
	 newline i)

(* each state in initial state is a marked place *)
(* ##### also prints other states in case no transition refers to it. wise ? ##### *)
and print_StatesDecl (a,b) =
    List.app (fn s => (print "pl "; print s; if mem s b then print " (1)" else print " (0)"; print "\n")) a

and print_Interval (left,right) =
    let fun printb b = print (Int.toString (Real.floor (b * (pow (10.0,!DIGITS)))))
	val Z = ref false
    in case left of
	   Open b => (print " ]"; printb b)
	 | Closed b => (Z := b <= 0.0; print " ["; printb b)
	 | Infinite => print " ]w";
       print ",";
       case right of
	   Open b => (printb b; print "[")
	 | Closed b => (printb b; print "]")
	 | Infinite => if !AADL andalso !Z then print "0]" else print "w["
    end

and print_Transition i (([tag],lab,_),src,c,interv,_) =
    let val tgt = get_target c
	val (read,inhib,from,tg) =  (* inhib for probes, in case needed *)
	    if !AADL then
		(* note: assumes no loop in fiacre source *)
		let val (rd,fr) = List.partition (fn p => mem p tgt) src
		in (map (fn s => s ^ "?") rd,
		    [],
		    fr,
		    List.filter (fn p => not (mem p rd)) tgt)
		end
	    else
		let val (rd,src') = List.partition (fn p => String.sub(p,String.size p - 1) = #"?") src
		    val (inh,fr) = List.partition (fn p => String.sub(p,String.size p - 2) = #"?") src'  (* ends with "?-" *)
		in (rd,
		    inh,
		    fr,
		    tgt)
		end
    in newline i; print "tr "; print tag; print " "; case lab of "" => () | _ => (print ": "; print lab);
       print_Interval interv;
       print_list (" "," ","") (fn p => print (p ^ "1")) read;
       print_list (" "," ","") (fn p => print (p ^ "1")) inhib;
       print_list (" "," ","") print from;
       print " -> ";
       print_list (""," ","") print tg
    end

(* gets target state *)
and get_target st =
    case st of
	StSequence(stm1,stm2) => get_target stm1 @ get_target stm2
      | StTo stl => stl
      | _ => []

and print_prior i (_,_,_,_,[]) = ()
  | print_prior i (([tg],_,_),_,_,_,prl) = (newline i; print "pr "; print tg; print " > "; print_list (""," ","") print prl)

fun emitFiacre s0 = (print_Fiacre s0; newline 0; flush ()) handle BYE => raise BYE | _ => failwith "emit net";

end; (* structure TtsNetEmit *)



(* prints the TTS C code by Prn *)
structure TtsCEmit = struct

open OutPrint;

fun newline n = if n<0 then () else (print "\n"; repeat n print " ");

(* type environment *)
val TYPES = ref ([] : (string * Type) list);
fun getBoundType s = assoc s (!TYPES);

(* holds temporary variables *)
(* #### could be avoided if temp variables has a specific constructor introduced by simplifier (e.g. TmpVarExp) #### *)
val TEMP = ref ([] : string list);

(* for tmp variable generation *)
(* q, e, r, a, c used in postcompose, with their own numbering; avoid them here *)
(* j, z, used here; i avoided because i386 is typically used in library headers ... *)
val new = let val UNIQUE = ref 0
	  in fn s => let val u = !UNIQUE
		     in UNIQUE := 1 + !UNIQUE;
			let val v = s ^ Int.toString u
			in TEMP := v :: !TEMP;
			   v
			end
		     end
	  end;

(* Returns the smallest C type that can store len elements *)
fun cTypeForLength len = if len <= 255 then
			 	UnsignedCharType
			 else if len <= 65535 then
				UnsignedShortType
			 else
				NatType

(* Returns the equivalent of the C sizeof() operator *)
fun sizeOf BoolType           = 1 (* Converted to a C unsigned char *)
|   sizeOf NatType            = 4
|   sizeOf IntType            = 4
|   sizeOf CharType           = 1
|   sizeOf UnsignedCharType   = 1
|   sizeOf ShortType          = 2
|   sizeOf UnsignedShortType  = 2
|   sizeOf (UnionType cs) =
    let val max = foldr (fn ((_, SOME ty), max) => Int.max (sizeOf ty,max) | (_, max) => max) 0 cs
    in max + sizeOf (cTypeForLength (length cs))
    end
|   sizeOf (EnumType names)      = sizeOf (cTypeForLength (List.length names))
|   sizeOf (QueueType (size,ty)) = sizeOf ty * size + (sizeOf (cTypeForLength (size)))
|   sizeOf (RecordType types)    = foldr (fn ((_,ty),acc) => sizeOf ty + acc) 0 types
|   sizeOf (ArrayType (size,ty)) = sizeOf ty * size
|   sizeOf (NamedType tyName) = sizeOf (getBoundType tyName);

(* Returns the equivalent of the C sizeof() operator, but taking into account that parts of values can be factorized *)
(* We consider that a type is factorized if its size is greater than the sum of
        - The size of the pointer needed for the indirection in the unicity table
        - The size of a cell in the AVL that stores the factorized data: two pointers + height (stored as an unsigned char)
          (it's very conservative: it's for the worst case where data can never be factorized)
   To be conservative, we assume that pointers are 64 bits.
   Thus, the final minimal size required is:
        1 * pointer + 2 * pointer + unsigned char

   Furthermore, we need to add the size of the type chosen to encode the length of the compressed factorized data. Again, to be conservative,
   we consider the maximum encoding type (e.g. unsigned int)
*)

val ptrSize = 8;
val factorizationThreshold = 3 * ptrSize + 1 + sizeOf NatType;

(* Store if a type is shared or not. Used as a cache. *)
val SHARED = ref ([] : string list);
fun shared n = mem n (!SHARED);

(* determines if a datatype is shared or not *)
(* never applied to a named type *)
(* "shared" check is done "bottom up", so named types found have been checked yet *)
fun shareable ty =
    (case ty of
	 RecordType [_] => false
       | RecordType types =>
	 foldr (fn ((_,ty),acc) => shSizeOf ty + acc) 0 types >= factorizationThreshold
       | ArrayType (size,nty) =>
	 if case nty of BoolType => size <= 32 | _ => false then
	     false
	 else
	     shSizeOf nty * size >= factorizationThreshold
       |  _ => false)
and shSizeOf ty =
	    case ty of
		BoolType            => 1 (* Converted to a C unsigned char *)
	      | NatType             => 4
	      | IntType             => 4
	      | CharType            => 1
	      | UnsignedCharType    => 1
	      | ShortType           => 2
	      | UnsignedShortType   => 2
	      | UnionType cs        =>
		let val max = foldr (fn ((_, SOME ty), max) => Int.max(shSizeOf ty,max) |   (_, max) => max) 0 cs
		in max + (shSizeOf (cTypeForLength (length cs)))
		end
	      | EnumType names      => shSizeOf (cTypeForLength (List.length names))
	      | QueueType (size,ty) => (* queues no more shared *)
		shSizeOf ty * size + (shSizeOf (cTypeForLength (size)))
	      | RecordType [(_, ty)] => shSizeOf ty
	      | RecordType types     => foldr (fn ((_,ty),acc) => shSizeOf ty + acc) 0 types
	      | ArrayType (size,ty) =>
		if case ty of BoolType => size <= 32 | _ => false then
		    if size <= 8  then
			shSizeOf UnsignedCharType
		    else if size <= 16 then
			shSizeOf UnsignedShortType
		    else
			shSizeOf NatType
		else shSizeOf ty * size
	      | NamedType tyName    => 
		if shared tyName then ptrSize else shSizeOf (getBoundType tyName);

(* assuming SHARED has been computed: *)
fun sharedSizeOf ty = if !SHARING then shSizeOf ty else sizeOf ty;

(* printing header file *)
fun print_Headers (Program ((_,dl,_)::_)) =
    let fun print_dec (TypeDecl (a,ty)) = (print "\n\n"; print_Type ty a; print ";")
	  | print_dec _ = ()
    in newline 0; print "/* fiacre headers */";
       newline 0; print "#include \"fiacre.h\"";
       List.app print_dec dl; newline 0; newline 0
    end

(* printing code file *)
and print_Fiacre (Program ((_,dl,d)::_)) =
    let fun print_dec d =
	    case d of
		FunctionDecl _ => (newline 0; print_FunctionDecl d)
	      | ExternDecl _ => (newline 0; print_ExternDecl d)
	      | TypeDecl (s,ty) => print_TypeDecl d
	      | _ => ()
    in newline 0; newline 0; print "static int avail;    /* for sprint_* */";
       newline 0; print "static char tmpbuff[16]; /* for sprint_* */";
       List.app print_dec dl;
       let val ProcessDecl (a,_,_,_,rk,k,f,_,_,_,j) = d
       in newline 0; newline 0; print "/* Process "; print a; print " */";
	  TEMP := map (fn (v,_,_) => v) (rk @ k);
	  if null rk then () else (newline 0; newline 0; print "/* read-only temporary variables */");
	  List.app (fn d => (newline 0; print_VarDecl d)) rk;
	  newline 0;
	  (* creates value type, compare_value, sprint_value: *)
	  newline 0; print "/* state type */";
	  let val valuetype = RecordType (map (fn (y,ty,_) => (y,ty)) f)
	  in print "\n\n"; print_Type valuetype "value"; print ";";
	     print_TypeDecl (TypeDecl ("value",valuetype))  (* functions *)
	  end;
	  print_initialFunc f rk;
	  print_transFunc j k;
	  print_indepFunc f;
	  print_obsFunc f
       end;
       newline 0
    end

and print_FunctionDecl (FunctionDecl (f,args,ty,vars,body)) =
    (print "\n/* Fiacre function "; print f; print " */";
     newline 0; print_Type ty ("fcr_"^f);
     print_list (" (",", ",") {") (fn (x,ty) => print_Type ty x) args;
     if null vars then () else (newline 4; print_list ("","; ","; ") (fn (x,ty,_) => print_Type ty x) vars);
     print_Statement 4 body "";
     newline 0; print "}")

and print_ExternDecl (ExternDecl (f,argtypes,ty)) =
    (newline 0;
     print "/* C function "; print f; print " */\n";
     print "extern "; print_Type ty f; print_list (" (",",",");") (fn ty => print_Type ty "") argtypes)
    
(* emit type declarations for named types *)
and print_TypeDecl (TypeDecl (a,ty)) =
    ((* print "\n\n"; print_Type ty a; print ";" *)  (* now printed in header file *)
     print "\n\n/* utilities for structure "; print a; print " */";
     print_compareFunc (a,ty);
     print_sprintFunc (a,ty);
     print_packFunc (a,ty))

and print_packFunc (typeName, ty) =
let
	val isShared = shared typeName
	val size = (sharedSizeOf ty) * 2 (* worst-case scenario *)
	val encodingTmp = cTypeForLength size
	val sharedLenEncoding = case cTypeForLength (size + (sizeOf encodingTmp)) of
		            	  UnsignedCharType  => "unsigned char"
		          	| UnsignedShortType => "unsigned short"
		          	| NatType           => "unsigned int"
	
	fun pack container varName (NamedType typeName) = 
	let
		val address = case getBoundType typeName of
				CharType          => ""
			      |	UnsignedCharType  => ""
			      |	ShortType         => ""	
			      | UnsignedShortType => ""
			      | IntType           => ""
			      | NatType  	  => ""
                              | BoolType          => ""
			      | EnumType _        => ""
			      | _                 => "&"
	in
	    if shared typeName then
		print ("pack_shared_" ^ typeName ^ "(" ^ address ^ container ^ varName ^ ", data)")
	    else
		print ("pack_" ^ typeName ^ "(" ^ address ^ container ^ varName ^ ", data)")
	end
	|   pack container name BoolType = print ("pack_bool(" ^ container ^ name ^ ", data)")
	|   pack container name CharType = print ("pack_char(" ^ container ^ name ^ ", data)")
	|   pack container name UnsignedCharType = print ("pack_unsigned_char(" ^ container ^ name ^ ", data)")
	|   pack container name ShortType = print ("pack_short(" ^ container ^ name ^ ", data)")
	|   pack container name UnsignedShortType = print ("pack_unsigned_short(" ^ container ^ name ^ ", data)")
	|   pack container name IntType = print ("pack_int(" ^ container ^ name ^ ", data)")
	|   pack container name NatType = print ("pack_unsigned_int(" ^ container ^ name ^ ", data)")

	fun unpack container varName (NamedType typeName) =
	    if shared typeName then
		print ("unpack_shared_" ^ typeName ^ "(&(" ^ container ^ varName ^ "), data)")
	    else
		print ("unpack_" ^ typeName ^ "(&(" ^ container ^ varName ^ "), data)")
	|   unpack container name BoolType = print ("unpack_bool(&(" ^ container ^ name ^ "), data)")
	|   unpack container name CharType = print ("unpack_char(&(" ^ container ^ name ^ "), data)")
	|   unpack container name UnsignedCharType = print ("unpack_unsigned_char((unsigned char*)&(" ^ container ^ name ^ "), data)")
	|   unpack container name ShortType = print ("unpack_short(&(" ^ container ^ name ^ "), data)")
	|   unpack container name UnsignedShortType = print ("unpack_unsigned_short((unsigned short*)&(" ^ container ^ name ^ "), data)")
	|   unpack container name IntType = print ("unpack_int(&(" ^ container ^ name ^ "), data)")
	|   unpack container name NatType = print ("unpack_unsigned_int(&(" ^ container ^ name ^ "), data)")

	fun printSharedFunsStr lenEnc =
	let
		val storageName = typeName ^ "_storage"
	in
		newline 0;
		newline 0; print "#ifdef SHARE";
		newline 0; print ("void pack_with_size_" ^ typeName ^ " (struct " ^ typeName ^ "* x, char* data) {");
		newline 2; print (lenEnc ^ " size = sizeof(" ^ lenEnc ^ ");");
		newline 2; print ("size += pack_" ^ typeName ^ "(x,data + size);");
		newline 2; print ("*(" ^ lenEnc ^ " *)data = size;");
		newline 0; print ("}");
		newline 0;
		newline 0; print ("void unpack_with_size_" ^ typeName ^ " (struct " ^ typeName ^ "* x, char* data) {");
		newline 2; print ("unpack_" ^ typeName ^ "(x,data + sizeof(" ^ lenEnc ^ "));");
		newline 0; print ("}");
		newline 0;
		newline 0; print ("size_t size_" ^ typeName ^ " (char* data) {");
		newline 2; print ("return *(" ^ lenEnc ^ " *)data;");
		newline 0; print ("}");
		newline 0;
		newline 0; print ("#define pack_shared_" ^ typeName ^ "(x,data)   pack_key(&" ^ typeName ^ "_storage,x,data)");
		newline 0; print ("#define unpack_shared_" ^ typeName ^ "(x,data) unpack_key(&" ^ typeName ^ "_storage,x,data)");
		newline 0;
		newline 0; print "#else";
		newline 0; print ("#define pack_shared_" ^ typeName ^ "(x,data)   pack_" ^ typeName ^ "(x,data)");
		newline 0; print ("#define unpack_shared_" ^ typeName ^ "(x,data) unpack_" ^ typeName ^ "(x,data)");
		newline 0; print "#endif"
	end

in
	case String.sub(typeName,0) of
		#"e" =>
		let
			val EnumType elems = ty
			val encoding = cTypeForLength (length elems)
		in
			newline 0;
			newline 0; print ("static size_t pack_" ^ typeName ^ " (enum " ^ typeName ^ " e, char* data) {");
			newline 0; print "  return "; pack "" "e" encoding; print ";";
			newline 0; print "}";
			newline 0;
			newline 0; print ("static size_t unpack_" ^ typeName ^ " (enum " ^ typeName ^ "* e, char* data) {");
			newline 0; print "  unsigned char tmp;";
			newline 0; print "  const size_t res = "; unpack "" "tmp" encoding; print ";";
			newline 0; print "  *e = tmp;";
			newline 0; print "  return res;";
			newline 0; print "}"
		end

	|	#"u" =>
		let
			val UnionType cs = ty
			val lenEncoding = cTypeForLength (List.length cs)
		in
			newline 0;
			newline 0; print ("static size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* u, char* data) {");
			newline 0; print "  const char* orig = data;";
			newline 0; print "  data += "; pack "u->" "con" lenEncoding; print ";";
			newline 0; print "  switch (u->con) {";
			List.app (fn (n, SOME ty) =>  (newline 4;print ("case " ^ n ^ ": data += "); pack  "u->arg." n ty; print "; break;")
				 |   (_,NONE) => ()
				 )
				 cs;
			newline 0; print "    default: break;";
			newline 0; print "  }";
			newline 0; print "  return data - orig;";
			newline 0; print "}";
			newline 0;
			newline 0; print ("static size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* u, char* data) {");
			newline 0; print "  const char* orig = data;";
			newline 0; print "  unsigned char tmp;";
 			newline 0; print "  data += "; unpack "" "tmp" lenEncoding; print ";";
			newline 0; print "  u->con = tmp;";
			newline 0; print "  switch (u->con) {";
			List.app (fn (n, SOME ty) => (newline 4;print ("case " ^ n ^ ": data += "); unpack  "u->arg." n ty; print "; break;")
				 |    (_, NONE) => ()
				 )
				 cs;
			newline 0; print "    default: break;";
			newline 0; print "  }";
			newline 0; print "  return data - orig;";
			newline 0; print "}"
		end

	|	#"q" =>
		let
			val QueueType(size,eltType) = ty
		in
			newline 0;
			newline 0; print ("static size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* q, char* data) {");
			newline 0; print "  const char* orig = data;";
			newline 0; print "  int i = 0;";
			newline 0; print "  data += pack_unsigned_int(q->len, data);";
			newline 0; print "  for( i = 0; i < q->len; ++i) {";
			newline 0; print "    data += "; pack "q->" "at[i]" eltType; print";";
			newline 0; print "  }";
			newline 0; print "  return data - orig;";
			newline 0; print "}";
			newline 0;
			newline 0; print ("static size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* q, char* data) {");
			newline 0; print "  const char* orig = data;";
			newline 0; print "  int i = 0;";
			newline 0; print "  data += unpack_unsigned_int(&q->len, data);";
			newline 0; print "  for( i = 0; i < q->len; ++i) {";
			newline 0; print "    data += "; unpack "q->" "at[i]" eltType; print";";
			newline 0; print "  }";
			newline 0; print "  return data - orig;";
			newline 0; print "}"
		end

	|	#"a" =>
		let
			val ArrayType(size,eltType) = ty
		in
			if case eltType of BoolType => size <= 32 | _ => false then
			let
			    val enc = if size <= 8 then "unsigned char" else if size <= 16 then "unsigned short" else "unsigned int"
			in 
			    newline 0;
			    newline 0; print ("static size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* a, char* data) {");
			    newline 0; print ("  size_t i; " ^ enc ^ " t = 0; ");
			    newline 0; print ("  for (i = 0; i < " ^ Int.toString(size) ^ "; ++i) {");
			    newline 0; print  "    t |= a->at[i] << i;";
			    newline 0; print  "  }";
			    newline 0; print ("  memcpy(data, &t, sizeof(" ^ enc ^ "));");
			    newline 0; print ("  return sizeof(" ^ enc ^ ");");
			    newline 0; print  "}";
			    newline 0;
			    newline 0; print ("static size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* a, char* data) {");
			    newline 0; print ("  size_t i; " ^ enc ^ " t = 0;");
			    newline 0; print ("  memcpy(&t, data, sizeof(" ^ enc ^ "));");
			    newline 0; print ("  for (i = 0; i < " ^ Int.toString(size) ^ "; ++i) {");
			    newline 0; print  "    a->at[i] = t & 0x01;";
			    newline 0; print  "    t >>= 1;";
			    newline 0; print  "  }";
			    newline 0; print ("  return sizeof(" ^ enc ^ ");");
			    newline 0; print  "}"
			end
			else
			(	newline 0;
				newline 0; print ("static size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* a, char* data) {");
				newline 0; print  "  const char* orig = data;";
				newline 0; print  "  size_t i = 0;";
				newline 0; print ("  for (i = 0; i < " ^ Int.toString(size) ^ "; ++i) {");
				newline 0; print  "    data += "; pack "a->" "at[i]" eltType; print ";";
				newline 0; print  "  }";
				newline 0; print "  return data - orig;";
				newline 0; print  "}";
				newline 0;
				newline 0; print ("static size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* a, char* data) {");
				newline 0; print  "  const char* orig = data;";
				newline 0; print  "  size_t i = 0;";
				newline 0; print ("  for (i = 0; i < " ^ Int.toString(size) ^ "; ++i) {");
				newline 0; print  "    data += "; unpack "a->" "at[i]" eltType; print ";";
				newline 0; print  "  }";
				newline 0; print  "  return data - orig;";
				newline 0; print  "}";
				if isShared then
				    printSharedFunsStr sharedLenEncoding
				else
				    ()
			)
		end

	|       #"r" =>
		let
			val RecordType types = ty
		in
			newline 0;
			newline 0; print ("static size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* r, char* data) {");
			newline 0; print  "  const char* orig = data;";
			List.app (fn (name,ty) => (newline 2; print "data += "; pack "r->" name ty; print ";")) types;
			newline 0; print  "  return data - orig;";
			newline 0; print  "}";
			newline 0;
			newline 0; print ("static size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* r, char* data) {");
			newline 0; print  "  const char* orig = data;";
			List.app (fn (name,ty) => (newline 2; print "data += "; unpack "r->" name ty; print ";")) types;
			newline 0; print  "  return data - orig;";
			newline 0; print  "}";
			if isShared then
			    printSharedFunsStr sharedLenEncoding
			else
			    ()
		end

	|       #"v" =>
		let
		    val (shNeedHash,shLenEncoding) =
			let val tySize = sharedSizeOf ty * 2 (* *2 => worst-case compression scenario*)
			    val needHash = tySize > 255
			    val size = if needHash then tySize + 4 (* 4 bytes to store an hash value *) else tySize
			    val encodingTmp = cTypeForLength size
			in (tySize > 256,
			    case cTypeForLength (size + (sizeOf encodingTmp)) of
				UnsignedCharType  => "unsigned char"
			      | UnsignedShortType => "unsigned short"
			      | NatType           => "unsigned int")
			end
		    val (needHash,lenEncoding) =
			let val tySize = sizeOf ty * 2 (* *2 => worst-case compression scenario*)
			    val needHash = tySize > 256
			    val size = if needHash then tySize + 4 (* 4 bytes to store an hash value *) else tySize
			    val encodingTmp = cTypeForLength size
			in (tySize > 256,
			    case cTypeForLength (size + (sizeOf encodingTmp)) of
				UnsignedCharType  => "unsigned char"
			      | UnsignedShortType => "unsigned short"
			      | NatType           => "unsigned int")
			end
		    val RecordType types = ty
		    val someshared = (* avoid SHARE if no change in size/hash whether SHARE of not *)
			not (null (!SHARED)) andalso (lenEncoding <> shLenEncoding orelse needHash <> shNeedHash)
		in
			(* values always stored with their size and possibly hash value, but size depends on SHARE *)

		        newline 0;
			newline 0; print ("size_t pack_" ^ typeName ^ " (struct " ^ typeName ^ "* v, char* data) {");
			newline 0; print ("  const char* orig = data;");
			List.app (fn (name,ty) => (newline 2; print "data += "; pack "v->" name ty; print ";")) types;
			newline 0; print ("  return data - orig;");
			newline 0; print "}";
			newline 0;

			newline 0; print ("size_t unpack_" ^ typeName ^ " (struct " ^ typeName ^ "* v, char* data) {");
			newline 0; print ("  const char* orig = data;");
			List.app (fn (name,ty) => (newline 2; print "data += "; unpack "v->" name ty; print ";")) types;
			newline 0; print ("  return data - orig;");
			newline 0; print "}";
			newline 0;

			newline 0;
			newline 0; print ("void pack_with_size_" ^ typeName ^ " (struct " ^ typeName ^ "* v, char* data) {");
			let fun p () =
				if needHash then
				    (newline 0; print ("  " ^ lenEncoding ^ " size = sizeof(int) + sizeof(" ^ lenEncoding ^ ");"))
				else
				    (newline 0; print ("  " ^ lenEncoding ^ " size = sizeof(" ^ lenEncoding ^ ");"))
			in if someshared then
			       (newline 0; print "#ifdef SHARE";
				if shNeedHash then
				    (newline 0; print ("  " ^ shLenEncoding ^ " size = sizeof(int) + sizeof(" ^ shLenEncoding ^ ");"))
				else
				    (newline 0; print ("  " ^ shLenEncoding ^ " size = sizeof(" ^ shLenEncoding ^ ");"));
				newline 0; print "#else";
				p ();
				newline 0; print "#endif")
			   else
			       p ()
			end;
			newline 0; print "  size += pack_value(v,data + size);";
			let fun p () =
				if needHash then
				    (newline 0; print ("  *(" ^ lenEncoding ^ " *)(data + sizeof(int)) = size;");
				     newline 0; print ("  MurmurHash3_x86_32(data + sizeof(int), size - sizeof(int), 42, (void*)data);"))
				else
				    (newline 0; print ("  *(" ^ lenEncoding ^ " *)data = size;"))
			in if someshared then
			       (newline 0; print "#ifdef SHARE";
				if shNeedHash then
				    (newline 0; print ("  *(" ^ shLenEncoding ^ " *)(data + sizeof(int)) = size;");
				     newline 0; print ("  MurmurHash3_x86_32(data + sizeof(int), size - sizeof(int), 42, (void*)data);"))
				else
				    (newline 0; print ("  *(" ^ shLenEncoding ^ " *)data = size;"));
				newline 0; print "#else";
				p ();
				newline 0; print "#endif")
			   else
			       p ()
			end;
			newline 0; print "}";

			newline 0;
			newline 0; print ("void unpack_with_size_" ^ typeName ^ " (struct " ^ typeName ^ "* v, char* data) {");
			let fun p () =
				if needHash then
				    (newline 0; print ("  unpack_value(v,data + sizeof(int) + sizeof(" ^ lenEncoding ^ "));"))
				else
				    (newline 0; print ("  unpack_value(v,data + sizeof(" ^ lenEncoding ^ "));"))
			in if someshared then
			       (newline 0; print "#ifdef SHARE";
				if shNeedHash then
				    (newline 0; print ("  unpack_value(v,data + sizeof(int) + sizeof(" ^ shLenEncoding ^ "));"))
				else
				    (newline 0; print ("  unpack_value(v,data + sizeof(" ^ shLenEncoding ^ "));"));
				newline 0; print "#else";
				p ();
				newline 0; print "#endif")
			   else
			       p ()
			end;
			newline 0; print "}";

			newline 0;
			newline 0; print "size_t size_value(char* data) {";
			let fun p () =
				if needHash then
				    (newline 0; print ("  return *(" ^ lenEncoding ^ " *)(data + sizeof(int));"))
				else
				    (newline 0; print ("  return *(" ^ lenEncoding ^ " *)(data);"))
			in if someshared then
			       (newline 0; print "#ifdef SHARE";
				if shNeedHash then
				    (newline 0; print ("  return *(" ^ shLenEncoding ^ " *)(data + sizeof(int));"))
				else
				    (newline 0; print ("  return *(" ^ shLenEncoding ^ " *)(data);"));
				newline 0; print "#else";
				p ();
				newline 0; print "#endif")
			   else
			       p ()
			end;
			newline 0; print "}"

		end

end

(* prints compare_a, sprint_a functions for type a *)
and print_compareFunc (a,ty) =
    let fun compare i (NamedType n) e1 e2 = (newline i; print ("{int r = compare_"^n^"("^e1^","^e2^"); if (r) return r;}"))
	  | compare i BoolType e1 e2 = (newline i; print ("if (!("^e1^") && ("^e2^")) return -1;");
					newline i; print ("if (("^e1^") && !("^e2^")) return 1;"))
	  | compare i _ e1 e2 = (newline i; print ("if (("^e1^") < ("^e2^")) return -1;");
				 newline i; print ("if (("^e1^") > ("^e2^")) return 1;"))
    in case String.sub(a,0) of
	   #"e" => (* enum *)
	   (newline 0; newline 0; print "static int compare_"; print a; print " (enum "; print a; print " e1, enum "; print a; print " e2) {";
	    newline 2; print "if (e1 < e2) return -1;";
	    newline 2; print "if (e1 > e2) return 1;";
	    newline 2; print "return 0;";
	    newline 0; print "}")
	 | #"u" => (* union *)
	   let val UnionType ctyl = getBoundType a
	       val zary = List.mapPartial (fn (c,NONE) => SOME c | _ => NONE) ctyl
	       val oary = List.mapPartial (fn (c,SOME ty) => SOME (c,ty) | _ => NONE) ctyl
	   in newline 0; print "\nstatic int compare_"; print a; print " (struct "; print a; print " e1, struct "; print a; print " e2) {";
	      newline 2; print ("if (e1.con < e2.con) return -1;");
	      newline 2; print ("if (e1.con > e2.con) return 1;");
	      newline 2; print ("switch (e1.con) {");
	      List.app (fn c => (newline 4; print "case "; print c; print " : {break;}")) zary;
	      List.app (fn (c,ty) => (newline 4; print "case "; print c; print " : {";
				      compare 6 ty ("e1.arg."^c) ("e2.arg."^c);
				      print " break;}")) oary;
	      newline 2; print "}";
	      newline 2; print "return 0;";
	      newline 0; print "}"
	   end
	 | #"q" => (* queue, only present items are compared *)
	   let val QueueType (sz,ety) = getBoundType a
	   in newline 0; print "\nstatic int compare_"; print a; print " (struct "; print a; print " e1, struct "; print a; print " e2) {";
	      compare 2 IntType "e1.len" "e2.len";
	      newline 2; print "{size_t i = 0;";
	      newline 4; print ("for (i=0; i< e1.len; i++) {");
	      compare 6 ety ("e1.at[i]") ("e2.at[i]");
	      newline 4; print "}";
	      newline 2; print "}";
	      newline 2; print "return 0;";
	      newline 0; print "}"
	   end
	 | #"a" => (* array *)
	   let val ArrayType (sz,ety) = getBoundType a
	   in newline 0; print "\nstatic int compare_"; print a; print " (struct "; print a; print " e1, struct "; print a; print " e2) {";
	      newline 2; print "size_t i = 0;";
	      newline 2; print ("for (i=0; i<"^Int.toString sz^"; i++) {");
	      compare 4 ety ("e1.at[i]") ("e2.at[i]");
	      newline 2; print "}";
	      newline 2; print "return 0;";
	      newline 0; print "}"
	   end
	 | #"r" => (* record *)
	   let val RecordType ftyl = getBoundType a
	   in newline 0; print "\nstatic int compare_"; print a; print " (struct "; print a; print " e1, struct "; print a; print " e2) {";
	      List.app (fn (f,ty) => compare 2 ty ("e1."^f) ("e2."^f)) ftyl;
	      newline 2; print "return 0;";
	      newline 0; print "}"
	   end
	 | #"v" => (* value type, a record *)
	   let val RecordType ftyl = ty
	   in newline 0; print "\nint compare_"; print a; print " (struct "; print a; print " *e1, struct "; print a; print " *e2) {";
	      List.app (fn (f,ty) => compare 2 ty ("e1->"^f) ("e2->"^f)) ftyl;
	      newline 2; print "return 0;";
	      newline 0; print "}"
	   end
    end

(* prints sprint_a function for type a *)
and print_sprintFunc (a,ty) =
     let fun sprint i (NamedType n) e = print ("{int r = sprint_"^n^" (buff,"^e^"); if (r<0) return r;} ")
	   | sprint i (BoolType) e = print ("if ("^e^") {CATI(4,\"true\");} else {CATI(5,\"false\");} ")
	   | sprint i (CharType) e = print ("CATV(\"%hhd\","^e^"); ")
	   | sprint i (UnsignedCharType) e = print ("CATV(\"%hhu\","^e^"); ")
	   | sprint i (ShortType) e = print ("CATV(\"%hd\","^e^"); ")
	   | sprint i (UnsignedShortType) e = print ("CATV(\"%hu\","^e^"); ")
	   | sprint i (NatType) e = print ("CATV(\"%u\","^e^"); ")
	   | sprint i _ e = print ("CATV(\"%d\","^e^"); ")
     in case String.sub(a,0) of
	    #"e" => (* enum *)
	    let val EnumType cl = ty
	    in newline 0; newline 0; print "static int sprint_"; print a; print " (char *buff, enum "; print a; print " e) {";
	       newline 2; print "switch (e) {";
	       List.app (fn c => (newline 4; print "case "; print c;
				  print " : {CATI("; printi (String.size c); print ",\""; print c; print "\"); break;}"))
			cl;
	       newline 2; print "}";
	       newline 2; print "return 0;";
	       newline 0; print "}"
	    end
	  | #"u" => (* union *)
	    let val UnionType ctyl = getBoundType a
		val zary = List.mapPartial (fn (c,NONE) => SOME c | _ => NONE) ctyl
		val oary = List.mapPartial (fn (c,SOME ty) => SOME (c,ty) | _ => NONE) ctyl
	    in newline 0; newline 0; print "static int sprint_"; print a; print " (char *buff, struct "; print a; print " e) {";
	       newline 2; print "switch (e.con) {";
	       List.app (fn (c,SOME ty) => (newline 4; print "case "; print c;
					    print " : {CATI("; printi (String.size c); print ",\""; print c; print "\"); ";
					    print "CATI(1,\"(\");"; sprint 6 ty ("e.arg."^c); print " CATI(1,\")\")";
					    print "; break;} ")
			  | (c,_) => (newline 4; print "case "; print c;
				      print " : {CATI("; printi (String.size c); print ",\""; print c; print "\"); break;} ")
			) ctyl;
	       newline 2; print "}";
	       newline 2; print "return 0;";
	       newline 0; print "}"
	    end
	  | #"q" => (* queue *)
	    let val QueueType(sz,ety) = ty
	    in newline 0; newline 0; print "static int sprint_"; print a; print " (char *buff, struct "; print a; print " e) {";
	       newline 2; print "{int qi;";
	       newline 4; print "CATI(2,\"{|\"); ";
	       newline 6; print ("for (qi=0;qi<e.len;qi++) {");
	       print "if (qi) {CATI(1,\",\");} ";
	       sprint 8 ety "e.at[qi]";
	       newline 6; print "}";
	       newline 4; print "CATI(2,\"|}\"); ";
	       newline 2; print "}";
	       newline 2; print "return 0;";
	       newline 0; print "}"
	     end
	  | #"a" => (* array *)
	    let val ArrayType(sz,ety) = ty
		val j = new "j"
	        val z = new "z"
	    in newline 0; newline 0; print "static int sprint_"; print a; print " (char *buff, struct "; print a; print " e) {";
	       newline 2; print "int "; print j; print "; ";
	       newline 2; print "int "; print z; print " = 0;";
	       newline 4; print "CATI(1,\"[\"); "; 
	       newline 4; print ("for ("^j^"=0;"^j^"<"^Int.toString sz^";"^j^"++) {");
	       newline 4; print ("if ("^z^") {CATI(1,\",\");} "^z^"++; ");
	       sprint 4 ety ("e.at["^j^"]");
	       newline 4; print "}";
	       newline 4; print "CATI(1,\"]\"); "; 
	       newline 2; print "return 0;";
	       newline 0; print "}"
	    end
	  | #"r" => (* record *)
	    let val RecordType ftyl = ty
		fun spr [(f,ty)] = (newline 2; print ("CATI(" ^ Int.toString (String.size f + 1) ^ ",\"" ^ f ^ "=" ^ "\"); ");
				    sprint 2 ty ("e." ^ f))
		  | spr (h::(t as _::_)) = (spr [h]; newline 2; print "CATI(1,\",\"); "; spr t)
	    in newline 0; newline 0; print "static int sprint_"; print a; print " (char *buff, struct "; print a; print " e) {";
	       newline 2; print "CATI(1,\"{\"); ";
	       spr ftyl;
	       newline 2; print "CATI(1,\"}\"); ";
	       newline 2; print "return 0;";
	       newline 0; print "}"
	     end
	  | #"v" => (* value type, a record; also initialize avail, avoid surrounding quotes *)
	    let val RecordType ftyl = ty
		fun spr [] = ()
		  | spr [(f,ty)] = (newline 2; print ("CATI(" ^ Int.toString (String.size f + 1) ^ ",\"" ^ f ^ "=" ^ "\"); ");
				    sprint 2 ty ("v->" ^ f))
		  | spr (h::(t as _::_)) = (spr [h]; newline 2; print "CATI(1,\",\"); "; spr t)
	    in newline 0; print "\nint sprint_state (int sz, char *buff, key s) {";
	       newline 2; print "struct value *v = lookup(s);";
	       newline 2; print "avail = sz;";
	       newline 2; print "strcpy(buff,\"\\0\");";
	       newline 2; print "/* prints state into buff: */";
	       (* newline 2; print "CATI(1,\"{\"); "; *)
	       spr ftyl;
	       (* newline 2; print "CATI(1,\"}\"); "; *)
	       newline 2; print "return strlen(buff);";
	       newline 0; print "}"
	     end
     end

and print_VarDecl (st,ty,_) = (print_Type ty st; print ";")

and print_initialFunc a b = (* a = state variables, b = rtmp variables *)
    let (* initializers have been statically evaluated, hence can only be "value" expressions, without variables *)
	fun print_Initializer st v stt =
	    case v of
		BoolExp _        => (newline 2; print stt; print st; print "="; print_Exp v ""; print";")
	      | IntExp _         => (newline 2; print stt; print st; print "="; print_Exp v ""; print";")
	      | EnumExp (a,_)    => (newline 2; print stt; print st; print "="; print a; print";")
	      | RecordExp (fl,_) => List.app (fn (f,l) => (print_Initializer (st^"."^f) l stt)) fl
	      | ArrayExp (el,_)  => List.app (fn (x,y) => print_Initializer (st^".at["^ Int.toString x  ^"]") y stt) 
					     (ListPair.zip(List.tabulate(List.length el,fn z => z),el)) 
	      | QueueExp (el,QueueType (z,ty)) =>
		let val len = List.length el
		in newline 2; print stt; print st; print ".len="; printi (length el); print";";
		   List.app (fn (x,y) => print_Initializer (st^".at["^ Int.toString x  ^"]") y stt) 
			    (ListPair.zip(List.tabulate(len,fn z => z),el))
		end
	      | ConExp (c,ty) =>
		(newline 2; print stt; print st; print ".con="; print c; print";")
	      | ConstrExp (c,e,_) =>
		(newline 2; print stt; print st; print ".con="; print c; print";";
		 print_Initializer (st ^ ".arg." ^ c) e stt)
	      | AppExp _ => (newline 0; print stt; print st; print "="; print_Exp v ""; print";")
	      | CallExp _ => (newline 0; print stt; print st; print "="; print_Exp v ""; print";")
	      | _ => failwith "unexpected initializer"

	in
		newline 0; newline 0; print "/* initializes temporaries and creates initial state */";
		newline 0; print "key initial() {";
		newline 2; print "struct value start;";
		(* initialize temporaries: *)
		if null b then () else (newline 2; print "/* read-only temporaries: */");
		List.app (fn x as (st,_,v) => print_Initializer st v "") b;
		(* initialize state variables: *)
		newline 2; print "/* state variables: */";
		List.app (fn x as (st,_,v) => print_Initializer st v "start.") a;
		if !PROPERTIES then
		    (* initialize auxiliary value variables *)
		    List.app (fn (_,v,e) => (newline 2; print "start."; print v; print " = "; print_Exp e "start."; print ";")) (!VALUEVARS)
		else
		    ();
		newline 2; print "/* storage initializations: */";
		newline 2; print "open_storage();";
		newline 2; print "values_storage = init_storage(sizeof(struct value), compare_value, pack_with_size_value, unpack_with_size_value, size_value);";
		case !SHARED of
		    [] => ()
		  | sh => (newline 0; print "#ifdef SHARE";
			   List.app (fn name =>
					(newline 2;
					 print (name ^ "_storage = init_storage(sizeof(struct " ^ name ^ ")"
						^                        ", compare_" ^ name
						^                        ", pack_with_size_" ^ name
						^                        ", unpack_with_size_" ^ name
						^                        ", size_" ^ name
						^                        ");")))
				    sh;
			   newline 0; print "#endif");
		newline 2; print "return store(&start);"; 
		newline 0; print "}";
		newline 0;
		newline 0; print "int final(int debug) {";
		newline 2; print ("finalize_storage(\"top\", debug, &values_storage);");
		case !SHARED of
		    [] => ()
		  | sh => (newline 0; print "#ifdef SHARE";
			   List.app (fn name => (newline 2; print ("finalize_storage(\"" ^ name ^ " \", debug, &" ^ name ^ "_storage);"))) sh;
			   newline 0; print "#endif");
		newline 0; print "  return close_storage(debug);";
		newline 0; print "}"
	end

and print_transFunc a rk =
    (newline 0; print "\n/* -- predicates and actions */";
     newline 0; print "/* by default, pre_t* returns true and act_t* is the identity function */";
     newline 0; newline 0; print "/* translation table for transition names */";
     newline 0; print ("char *transtable["^ Int.toString(List.length a) ^"] = "); 
     print_list ("{",",","}") (fn (([st],_,_),_,_,_,_) => print ("\""^ hd (String.fields (fn #":" => true | _ => false) st)^"\"")) a; print "; ";
     newline 0; print "\nchar **transitions (int *sz) {";
     newline 0; print "*sz ="; print (Int.toString(List.length a)); print "; ";
     newline 0; print "return transtable;";
     newline 0; print "}";
     List.app (fn (i,(([tag],lab,_),_,stm,_,_)) => print_preFunc (tag^(case lab of "" => "" | _ => (":"^lab))) i stm [] rk)
	      (ListPair.zip(List.tabulate(List.length a,fn y=>y),a));
     List.app (fn (i,(([tag],lab,_),_,stm,_,_)) => print_actFunc (tag^(case lab of "" => "" | _ => (":"^lab))) i stm rk)
	      (ListPair.zip(List.tabulate(List.length a,fn y=>y),a)))

(* returns true if statement or expression updates some state variable *)
and assigns dl st =
    case st of
	StGuard e => assignse dl e
      | StSequence(stm1,stm2) => assigns dl stm1 orelse assigns dl stm2
      | StAssign(p,exp) => assignsp dl p orelse assignse dl exp
      | StWhile (e,s) => assignse dl e orelse assigns dl s
      | StForeach (v,s) => assignse dl v orelse assigns dl s
      | StCond (e,s1,s2) => assignse dl e orelse assigns dl s1 orelse assigns dl s2
      | StBlock (_,s) => assigns dl s
      | _ => false
and assignsp dl p =
    case p of
	VarExp (v,_) => (case List.find (fn (v',_) => v=v') dl of SOME _ => false | _ => not (mem v (!TEMP)))
      | ArrayAccessExp(a,i,_) => assignsp dl a orelse assignse dl i
      | RecordAccessExp(r,_,_) => assignsp dl r
and assignse dl e =
    matchExp (fn LetExp (s,e) => assigns dl s | _ => false) e

and print_preFunc tag index stm dl k =
    case stm of
	StBlock(dl',s) => print_preFunc tag index s (dl @ dl') k
      | StSequence(s,_) => print_preFunc tag index s dl k
      | StGuard e =>
	let fun print_unroll_Exp e prf = (* user friendly *)
		case e of
		    LetExp (s,e') => (print_Statement 2 s prf; print_unroll_Exp e' prf)
		  | InfixExp (AND,_,e1,e2) => (print_unroll_Exp e1 prf; print_unroll_Exp e2 prf)
		  | UnopExp(NOT,_,UnopExp(NOT,_,e)) => print_unroll_Exp e prf
		  | UnopExp(NOT,_,e) => (newline 2; print "if ("; print_Exp e prf; print ") return 0;")
		  | _ => (newline 2; print "if (!("; print_Exp e prf; print ")) return 0;")
	    fun print_guard dl e prf =
		if null dl
		then print_unroll_Exp e prf
		else (newline 2;
		      print "{"; List.app (fn (s,ty) => (TEMP := s :: !TEMP; newline 4; print_Type ty s; print "; ")) dl;
		      print_unroll_Exp e prf;
		      newline 2; print "}")
	in newline 0; newline 0; print ("bool pre_"^ Int.toString(index) ^" (key s) {");
	   newline 2; print "/* "; print tag; print " */";
	   (* if null k then () else (newline 2; print "/* temporaries */"); *)
	   List.app (fn d => (newline 2; print_VarDecl d)) k;
	   if assignse dl e
	   then (* #### improve: make local copies of state variables updated in the block, rather than all #### *)
	        (newline 2; print "struct value new = *(struct value *)(lookup(s));";
		 print_guard dl e "new.")
	   else (newline 2; print ("struct value *v = lookup(s);");
		 print_guard dl e "v->");
	   newline 2; print "return 1;"; 
	   newline 0; print "}"	   
	end
      | _ => (* no guard *) ()

and print_actFunc tag index stm k =
    if assigns [] stm
    then let fun hasGuard s = case s of StGuard _ => true | StBlock (_,s) => hasGuard s | _ => false
	     fun print_act s dl =
		 case s of
		     StBlock(dl',s) => print_act s (dl @ dl')
		   | StSequence(s1,s2) =>
		     if matchStm hasGuard s1
		     then print_act s2 dl
		     else (newline 0; newline 0; print ("key act_"^ Int.toString(index) ^" (key s) {");
			   newline 2; print "/* "; print tag; print " */";
			   (* if null k then () else (newline 2; print "/* temporaries */"); *)
			   List.app (fn d => (newline 2; print_VarDecl d)) k;
			   newline 2; print ("struct value new = *(struct value*)lookup(s);");
			   print_Statement 2 s "new.";
			   newline 2; print ("return store(&new);");
			   newline 0; print "}")
		   | _ => ()
	 in print_act stm []
	 end
    else ()

(* prints independent function *)
and print_indepFunc varL =
    (newline 0; newline 0; print "/* -- independance relation for transitions, for partial order modes and reset in timed modes -- */";
     newline 0; print "/* indices refer to transtable; specifies independance for ALL transitions */";
     newline 0; print "bool independant (int t1, int t2) {";
     newline 2; print "  /* always true */";
     newline 2; print "  return 1;";
     newline 0; print "}")

(* prints structure components in C style *)
and print_obsFunc varL =
    let	fun getVarNames st ty =
	    case ty of
		BoolType => [("(int)(v->"^st^")",st)]
	      | NatType => [("(int)(v->"^st^")",st)]
	      | IntType => [("(int)(v->"^st^")",st)]
	      | CharType => [("(int)(v->"^st^")",st)]
	      | UnsignedCharType => [("(int)(v->"^st^")",st)]
	      | ShortType => [("(int)(v->"^st^")",st)]
	      | UnsignedShortType => [("(int)(v->"^st^")",st)]
	      | EnumType _ => [("(int)(v->"^st^")",st)]
	      | UnionType cs => (* [("(int)(v->"^st^".con)",st)] *)
		let val con as (c,_) = ("(int)(v->"^st^".con)",st^".con")
		in con :: List.concat (List.map (fn (n,SOME ty) =>
						    let val ly = getVarNames (st^".arg."^n) ty
						    in map (fn (uu,vv) => ("("^c^" == "^n^") ? ("^uu^") : 0",vv)) ly
						    end
						  | (_,NONE) => []) cs)
		end
	      | RecordType stList => List.concat(List.map (fn (x,y) => getVarNames (st^"."^x) y) stList)
	      | ArrayType(s,tty) => List.concat(List.tabulate(s,fn y => getVarNames (st^".at["^ Int.toString(y)  ^"]") tty))
	      | QueueType (s,tty) =>
		let val len as (c,_) = ("(int)(v->"^st^".len)",st^".len")
		in len :: List.concat(List.tabulate(s,fn y => let val ly = getVarNames (st^".at["^ Int.toString(y)  ^"]") tty
							      in map (fn (uu,vv) => ("("^c^" > "^Int.toString y^") ? ("^uu^") : 0",vv)) ly
							      end))
		end
	      | NamedType aa => getVarNames st (getBoundType aa)
	val (str1,str2) = ListPair.unzip (flat (map (fn (stl,b,c) => getVarNames stl b) varL))
	fun encode n = (* surrounds n with braces if contains "." or "[" *) (* ### should suppress ".at" too ? ### *)
	    if null (List.filter (fn #"." => true | #"[" => true | _ => false) (String.explode n))
	    then "\"" ^ n ^ "\"" else "\"{" ^ n ^ "}\""
    in  newline 0; print "\n/* -- state observers for kts output -- */";
	newline 0; print ("char *obsnames[" ^ Int.toString(List.length str2) ^ "] = {");
	print_list ("",",","") (fn y => print (encode y)) str2; print "};";
	newline 0; print ("\nchar **obs_names (int *count) {*count = " ^ Int.toString(List.length str2)  ^"; return obsnames;}");
	newline 0; print ("\n/* returns values of observables for state s (abstractions for noninteger fields) */");
	newline 0; print ("int ovalues["^ Int.toString (List.length str2)  ^"];");
	newline 0; newline 0; print "int *obs_values (key s) {";
	newline 0; print "struct value *v = lookup(s);";
	case str2 of
	    [] => ()
	  | _ => (newline 0; print_list ("",";\n","; ")
					(fn (x,y) => print ("ovalues[" ^ Int.toString(x) ^ "]=" ^ y))
					(ListPair.zip(List.tabulate(List.length str1,(fn y=>y)),str1)));
	newline 0; print "return ovalues;";
	newline 0; print "}"
    end 

and print_Type ty a =
    let fun printn "" = () | printn a = (print " "; print a)
    in case ty of
	   BoolType => (print "bool"; printn a)
      | NatType  => (print "unsigned int"; printn a)
      | CharType => (print "char"; printn a)
      | UnsignedCharType => (print "unsigned char"; printn a)
      | ShortType => (print "short"; printn a)
      | UnsignedShortType => (print "unsigned short"; printn a)
      | IntType  => (print "int"; printn a)
      | NamedType nt =>	(case getBoundType nt of
			     EnumType _ => (print "enum "; print nt; printn a)
			   | UnionType _ => (print "struct "; print nt;  printn a) (* unions are records *)
			   | RecordType _ => (print "struct "; print nt;  printn a)
			   | QueueType _ => (print "struct "; print nt;  printn a) (* queues are records *)
			   | ArrayType _ => (print "struct "; print nt;  printn a)
			   | _ => failwith "print_Named_Type")
      | EnumType cl => (print "enum "; print a; print_list (" {",",","}") print cl)
      | UnionType ctyl => let val l = List.mapPartial (fn (c,SOME ty) => SOME (c,ty) | _ => NONE) ctyl
			  in print "struct "; print a; print " {";
			     newline 2; print "enum "; print_list (" {",",","}") (fn (c,_) => print c) ctyl; print " con;";
			     newline 2; print "union {";
			     List.app (fn (c,ty) => (newline 4; print_Type ty c; print ";")) l;
			     newline 4; print "int none;"; (* now useless ? *)
			     newline 2; print "} arg;";
			     newline 0; print "}"
			  end
      | RecordType rl => (print "struct "; print a;
			  print_list (" {","","") (fn (x,y) => (newline 2; print_Type y x; print "; ")) rl;
			  newline 0; print "}")
      | ArrayType (s,t) => (print "struct "; print a; print " {";
			    newline 2; print_Type t "at"; print "["; print (Int.toString s); print "];";
			    newline 0; print "}")			  
      | QueueType(s,b) => (print "struct "; print a; print " {";
			   newline 2; print "unsigned int len;";
			   newline 2; print_Type b "at"; print "["; print (Int.toString s); print "];";
			   newline 0; print "}")
      | _ => failwith "print_Type"
    end

(* prints statements in a act_* (without StGuard) *)
and print_Statement i stm stt =
    case stm of
        StSequence(stm1,stm2) => (print_Statement i stm1 stt; print_Statement i stm2 stt)
      | StCond(e,s1,StNull) => (newline i; print "if ("; print_Exp e stt; print ") {";
				print_Statement (i+2) s1 stt;
				newline i; print "}")
      | StCond(e,StNull,s2) => (newline i; print "if (!("; print_Exp e stt; print ")) {";
				print_Statement (i+2) s2 stt;
				newline i; print "}")
      | StCond(e,s1,s2) => (newline i; print "if ("; print_Exp e stt; print ") {";
			    print_Statement (i+2) s1 stt;
			    newline i; print "} else {";
			    print_Statement (i+2) s2 stt;
			    newline i; print "}")
      | StWhile (e,s) => (newline i; print "while ("; print_Exp e stt; print ") {";
			  print_Statement (i+2) s stt;
			  newline i; print "}")
      | StForeach (e as VarExp(v,IntervalType(fr,to)),s) => (newline i; print "for (";
							     print_Exp e stt; print "="; printi fr; print ";";
							     print_Exp e stt; print "<="; printi to; print ";";
							     print_Exp e stt; print "++) {";
							     print_Statement (i+2) s stt;
							     newline i; print "}")
      | StAssign(acc,exp) =>
	(* expansion must be applied recursively. e.g.  a := b when a : array 4 of array 8 of ... *)
	if acc=exp then () else
	(case exp of
	     LetExp (stm,e) => (print_Statement i stm stt; print_Statement i (StAssign(acc,e)) stt)
	   | RecordExp (fle,RecordType ftyl) =>
	     List.app (fn (f,e) => print_Statement i (StAssign(RecordAccessExp(acc,f,assoc f ftyl),e)) stt) fle
	   | ArrayExp (el,ArrayType (sz,ety)) =>
	     (List.tabulate (sz,fn j => print_Statement i (StAssign(ArrayAccessExp(acc,IntExp(j,IntType),ety),List.nth(el,j))) stt); ())
	   | QueueExp (el,QueueType(sz,ety)) =>
	     let val len = length el
		 val aty = ArrayType(sz,ety)
	     in print_Statement i (StAssign(RecordAccessExp(acc,"len",IntType),IntExp(len,IntType))) stt;
		List.app (fn j => let val s = StAssign(ArrayAccessExp(acc,IntExp(j,IntType),ety),List.nth(el,j))
				  in print_Statement i s stt
				  end)
			 (List.tabulate(len,fn k => k))
	     end
	   | BinopExp(ENQUEUE,QueueType(sz,ety),q,e) =>
	     (newline i; print "ENQUEUE("; print_Exp acc stt; print ","; print_Exp q stt; print ","; print_Exp e stt; print ","; printi sz; print "); ")
	   | BinopExp(APPEND,QueueType(sz,ety),q,e) => (* assign result then modify *)
	     (newline i; print "APPEND("; print_Exp acc stt; print ","; print_Exp q stt; print ","; print_Exp e stt; print ","; printi sz; print "); ")
	   | UnopExp(DEQUEUE,QueueType(sz,ety),q) =>
	     (newline i; print "DEQUEUE("; print_Exp acc stt; print ","; print_Exp q stt; print "); ")
	   | ConExp (c,ty) =>
	     (newline i; print_Exp acc stt; print ".con = "; print c; print ";")
	   | ConstrExp (c,e,UnionType ctyl) =>
	     (newline i; print_Exp acc stt; print ".con = "; print c; print "; ";
	      print_Statement i (StAssign(RecordAccessExp(acc,"arg." ^ c,let val SOME ty = assoc c ctyl in ty end),e)) stt)
 	   | _ => (newline i; print_Exp acc stt; print "="; print_Exp exp stt; print"; "))
      | StGuard exp => ()
      | StBlock (dl,ss) =>
	(newline i; print "{";
	 List.app (fn (s,ty) => (TEMP := s :: !TEMP; print_Type ty s; print "; ")) dl;
	 print_Statement i ss stt;
	 newline i; print "}")
      | StReturn exp => (newline i; print "return "; print_Exp exp ""; print ";")
      | StFailwith m => (newline i; print "FAILWITH (\""; print m; print "\");")
      | _ => ()

(* prints guards or rhs of assignements or initializer components *)
and print_Exp e stt =
    case e of
	VarExp (a,_)            => (if mem a (!TEMP) then () else print stt; print a)
      | RecordAccessExp (a,b,_) => (print_Exp a stt; print "."; print b)
      | ArrayAccessExp (a,b,_)  => (print_Exp a stt; print ".at["; print_Exp b stt; print "]")
      | IntExp (n,_)            => printi n
      | BoolExp b               => if b then print "1" else print "0" 
      | EnumExp (e,_)           => print e
      | DestrExp (c,e)          => (print_Exp e stt; print (".arg." ^ c))
      | MatchesExp (c,e)        => (print "("; print_Exp e stt; print ".con == "; print c; print ")")
      | InfixExp(x,ty,y,z) 	=>
	(case x of
	     ADD => print_arith_infix ty (y,z) "ADD" "+" stt
	   | SUB => print_arith_infix ty (y,z) "SUB" "-" stt
	   | MUL => print_arith_infix ty (y,z) "MUL" "*" stt
	   | DIV => print_arith_infix ty (y,z) "DIV" "/" stt
	   | MOD => print_arith_infix ty (y,z) "MOD" "%" stt
	   | EQ  => (case ty of
			 NamedType n => (print "(! compare_"; print n; print "("; print_Exp y stt; print ","; print_Exp z stt; print "))")
		       | _ => (print "("; print_Exp y stt; print " == "; print_Exp z stt; print ")"))
	   | NE  => (* if matches (y,z) then print "0" else *) (print "("; print "!"; print_Exp (InfixExp (EQ,ty,y,z)) stt; print ")")
	   | _ => let val f = case x of AND => "&&" | OR  => "||" | LE  => "<=" | LT  => "<" | GT  => ">" | GE  => ">="
		  in print "("; print_Exp y stt; print " "; print f; print " "; print_Exp z stt; print ")"
		  end)
      | UnopExp(a,ty,b)         =>
	(case a of
	     MINUS  => print_arith_unop ty b "MINUS" "-" stt
	   | PLUS   => print_Exp b stt
	   | COERCE => print_arith_unop ty b "COERCE" "" stt
	   | NOT    => (case b of
			    UnopExp(NOT,_,b') => print_Exp b' stt
			  | _ => (print "("; print "!"; print_Exp b stt; print ")"))
	   | _ =>
	     let val QueueType(sz,_) = ty
	     in case a of
		    EMPTY  => (print "EMPTY("; print_Exp b stt; print ")")
		  | FULL   => (print "FULL("; print_Exp b stt; print ","; printi sz; print ")")
		  | LENGTH => (print "LENGTH("; print_Exp b stt; print ")")
		  | FIRST  => (print "FIRST("; print_Exp b stt; print ")")
		  | _ => failwith "unexpected queue unopexp" (* dequeue *)
	     end)
      | CondExp(c,e1,e2)        => (print "(("; print_Exp c stt; print ") ? ("; print_Exp e1 stt; print ") : ("; print_Exp e2 stt; print "))")
      | LetExp(s,e)             => (print "({"; print_Statement (~1000) s stt; print_Exp e stt; print ";})")
      | AppExp (f,el,_,_)       => (print ("fcr_"^f); print_list ("(",",",")") (fn e => print_Exp e stt) el)
      | CallExp (f,el,_,_)      => (print f; print_list ("(",",",")") (fn e => print_Exp e stt) el)
      (* the following cases should not appear; normalized in postcompose *)
      | BinopExp(x,ty,y,z)	=> failwith "unexpected binop expression in print_Exp"
      | RecordExp _             => failwith "unexpected record expression in print_Exp"
      | ArrayExp _              => failwith "unexpected array expression in print_Exp"
      | QueueExp _              => failwith "unexpected queue expression in print_Exp"
      | ConExp (c,_)            => failwith ("unexpected 0-ary construction ("^c^") in print_Exp")
      | ConstrExp (c,_,_)       => failwith ("unexpected 1-ary construction ("^c^" ...) in print_Exp")
      | _                       => failwith "unexpected expression in print_Exp"

and print_arith_infix ty (y,z) f ope stt =
    case ty of
	IntervalType(l,h) => (print ("I" ^ f); print "("; print_Exp y stt; print ","; print_Exp z stt; print ","; printi l; print ","; printi h; print ")")
      | NatType => (print ("N" ^ f); print "("; print_Exp y stt; print ","; print_Exp z stt; print ")")
      | _ => (print "("; print_Exp y stt; print " "; print ope; print " "; print_Exp z stt; print ")")

and print_arith_unop ty x f ope stt =
    case ty of
	IntervalType(l,h) => (print ("I" ^ f); print "("; print_Exp x stt; print ","; printi l; print ","; printi h; print ")")
      | NatType => (print ("N" ^ f); print "("; print_Exp x stt; print ")")
      | _ => (print "("; print ope; print_Exp x stt; print ")")


(* emits .c iff some data *)
fun emitFiacre (hfile,fdef,writeh) (p as Program ((_,dl,d as ProcessDecl (_,_,_,_,rk,_,f,_,_,_,_))::_)) =
      (print ("/* Created " ^ Date.toString (Date.fromTimeLocal(Time.now ())) ^ " by frac version " ^ version ^ " */");
       newline 0; newline 0; print "#include <stdlib.h>";
       newline 0; print "#include <math.h>";
       newline 0; print "#include <string.h>";
       newline 0; print "#include <stdio.h>";
       newline 0; print "#include <stddef.h>";
       newline 0; print "\n/* tts API version */";
       newline 0; print "int ttsapiversion () {return 1;}";
       newline 0; print "\n/* provider */";
       newline 0;
       print "char *source_info[3] = {\"frac\",\"";
       print version; print "\",\"";
       print (Date.toString (Date.fromTimeLocal(Time.now ()))); print "\"};";
       print "\nchar **sourceinfo () {return source_info;}";
       flush ();
       if List.null f andalso List.null rk then
	   (newline 0; print "/* minimal functions */";
	    newline 0; print "typedef int bool;";
	    newline 0; print "typedef long key;";
	    newline 0; print "char buff[1];";
	    newline 0; print "int sprint_state (int sz, char *buff, key s) {strcpy(buff,\"\\0\"); return 0;}";
	    newline 0; print "key initial() {return 0;}";
	    newline 0; print "char **transitions (int *sz) {*sz =0; return 0;}";
	    newline 0; print "char **obs_names (int *count) {*count = 0; return 0;}";
	    newline 0; print "int *obs_values (key s) {return 0;}";
	    newline 0; print "bool independant (int t1, int t2) {return 1;}";
	    newline 0; print "int final(int debug) {return 0;}";
	    newline 0; flush ();
	    OS.FileSys.remove fdef handle _ => ())
       else
	   let val p' as Program ((_,dl',_)::_) = Name.nameTypes p
	   in (* Construct the list of types (in order of declaration) *)
	      TYPES := List.mapPartial (fn (TypeDecl(a,ty)) => SOME (a,ty) | _ => NONE) dl';
	      (* Compute the shared types bottom up (IMPORTANT!) *)
	      if !SHARING then
		  (List.app (fn (name,ty) => if shareable ty then SHARED := name :: !SHARED else ()) (!TYPES);
		   SHARED := rev (!SHARED))  (* not essential, just for code readability *)
	      else
		  ();
	      (* Do not share a type if it's alone at the top-level *)
	      case f of [(_, NamedType tyName, _)] => SHARED := List.filter (fn n => n <> tyName) (!SHARED) | _ => ();
	      writeh print_Headers p';
	      newline 0; print "\n/* fiacre and private headers */";
	      newline 0; print "#include \""; print hfile; print "\"";
	      newline 0; newline 0; print "/* storage headers */";
	      newline 0; print "#include \"pack.h\"";
	      newline 0; print "#include \"store.h\"";
	      newline 0;
	      newline 0; print "static storage_t values_storage;";
	      case !SHARED of
		  [] => ()
		| sh => (newline 0; print "#ifdef SHARE";
			 List.app (fn name => (newline 0; print ("static storage_t " ^ name ^ "_storage;"))) sh;
			 newline 0; print "#endif");
	      newline 0; print "#define lookup(s)   lookup_data(&values_storage,s)";
	      newline 0; print "#define store(v)    store_data(&values_storage,v)";
	      print_Fiacre p';
	      newline 0; flush ()
	   end)

       handle BYE => raise BYE | _ => failwith "emit c";
   
end; (* structure TtsCEmit *)


val ltllib = String.concat
[
"\n", 
"\n# general",
"\nop deadlockFree = [] - dead;", 
"\nop infinitelyOften L = [] <> L;",
"\nop mortal L = <> [] - L;",
"\n",
"\n",
"\n# patterns, source http://patterns.projects.cis.ksu.edu/documentation/patterns.shtml",
"\n",
"\n# Weak until",
"\ninfix 0 p W q = p U (q \\/ []p);",
"\n",
"\n# absence",
"\nop absent P = [] - P;",
"\nop absentBefore P R = <>R => (-P U R);",
"\nop absentAfter P Q = [](Q => [] - P);",
"\nop absentBetween P Q R = [] ((Q /\\ -R /\\ <>R) => (-P U R));",
"\nop absentUntil P Q R = [](Q /\\ -R => (-P W R));",
"\n",
"\n# presence",
"\nop present P = <> P;",
"\nop presentBefore P Q = -Q W (P /\\ -Q);",
"\nop presentAfter P Q = [] - Q \\/ <> (Q /\\ <>P);",
"\nop presentBetween P Q R = [] (Q /\\ -R => (-R W (P /\\ -R)));",
"\nop presentUntil P Q R = [] (Q /\\ -R => (-R U (P /\\ -R)));",
"\n",
"\n# universality",
"\nop always P = [] P;",
"\nop alwaysBefore P R = <> R => (P U R);",
"\nop alwaysAfter P Q = [] (Q => []P);",
"\nop alwaysBetween P Q R = [] ((Q /\\ -R /\\ <>R) => (P U R));",
"\nop alwaysUntil P Q R = [] (Q /\\ -R => (P W R));",
"\n",
"\n# precedence",
"\nop precedes S P = -P W S;",
"\nop precedesBefore S P R = <>R => (-P U (S \\/ R));",
"\nop precedesAfter S P Q = [] - Q \\/ <>(Q /\\ (-P W S));",
"\nop precedesBetween S P Q R = []((Q /\\ -R /\\ <>R) => (-P U (S \\/ R)));",
"\nop precedesUntil S P Q R = [](Q /\\ -R => (-P W (S \\/ R)));",
"\n",
"\n# response",
"\nop leadsto P S = [] (P => <>S);",
"\nop leadstoBefore P S R = <>R => (P => (-R U (S /\\ -R))) U R;",
"\nop leadstoAfter P S Q = [] (Q => [](P => <>S));",
"\nop leadstoBetween P S Q R = [] ((Q /\\ -R /\\ <>R) => (P => (-R U (S /\\ -R))) U R);",
"\nop leadstoUntil P S Q R = [] (Q /\\ -R => ((P => (-R U (S /\\ -R))) W R));",
"\n",
"\n"
];


val mmclib = String.concat
[
"\n", 
"\n# ctl",
"\nprefix AF g = min x | g \\/ (<T>T /\\ [T]x);",
"\nprefix EF g = min x | g \\/ <T>x;",
"\nprefix AG g = max x | g /\\ [T]x;",
"\nprefix EG g = max x | g /\\ <T>x;",
"\ninfix f AU g = min x | g \\/ (f /\\ <T>x /\\ [T]x);",
"\ninfix f EU g = min x | g \\/ (f /\\ <T>x);",
"\n", 
"\n# general",
"\nop deadlockFree = - <T>T;", 
"\nop live L = AG EF L;",
"\nop mortal L = AF EG - L;",
"\n",
"\n",
"\n# patterns, source http://patterns.projects.cis.ksu.edu/documentation/patterns.shtml",
"\n",
"\n# Weak AW",
"\ninfix f AW g = (-g) EU (-f /\\ -g);",
"\n",
"\n# absence",
"\nop absent P = AG - P;",
"\nop absentBefore P R = (-P \\/ AG - R) AW R;",
"\nop absentAfter P Q = AG (Q => AG - P);",
"\nop absentBetween P Q R = AG (Q /\\ -R => ((-P \\/ AG - R) AW R));",
"\nop absentUntil P Q R = AG (Q /\\ -R => (-P AW R));",
"\n",
"\n# presence",
"\nop present P = AF P;",
"\nop presentBefore P R = (- R) AW (P /\\ (- R));",
"\nop presentAfter P Q = -Q AW (Q /\\ AF P);",
"\nop presentBetween P Q R = AG (Q /\\ -R => -R AW (P /\\ -R));",
"\nop presentUntil P Q R =   AG (Q /\\ -R => -R AU (P /\\ -R));",
"\n",
"\n# universality",
"\nop always P = AG P;",
"\nop alwaysBefore P R = (P \\/ AG - R) AW R;",
"\nop alwaysAfter P Q = AG (Q => AG P);",
"\nop alwaysBetween P Q R = AG (Q /\\ -R => ((P \\/ AG - R) AW R));",
"\nop alwaysUntil P Q R = AG (Q /\\ -R => P AW R);",
"\n",
"\n# precedence",
"\nop precedes S P = -P AW S;",
"\nop precedesBefore S P R = (-P \\/ AG - R) AW (S \\/ R);",
"\nop precedesAfter S P Q = -Q AW (Q /\\ (-P AW S));",
"\nop precedesBetween S P Q R = AG (Q /\\ -R => (-P \\/ AG - R) AW (S \\/ R));",
"\nop precedesUntil S P Q R = AG (Q /\\ -R => (-P AW (S \\/ R)));",
"\n",
"\n# response",
"\nop leadsto P S = AG (P => AF S);",
"\nop leadstoBefore P S R = ((P => -R AU (S /\\ -R)) \\/ AG -R) AW R;",
"\nop leadstoAfter P S Q = -Q AW (Q /\\ AG (P => AF S));",
"\nop leadstoBetween P S Q R = AG (Q /\\ -R => ((P => -R AU (S /\\ -R)) \\/ AG - R) AW R);",
"\nop leadstoUntil P S Q R = AG (Q /\\ -R => (P => -R AU (S /\\ -R)) AW R);",
"\n",
"\n"
];


(* prints tl formulas by Prn *)
structure TtsLtlEmit = struct

open OutPrint;

fun newline n = if n<0 then () else (print "\n"; repeat n print " ");

fun printValue v =
    case v of
	VarExp (v,_) => print v
      | IntExp (i,_) => print (Int.toString i)
      | BoolExp b => print (if b then "1" else "0")
      | InfixExp (f,_,e1,e2) =>
	(case f of
	     AND => (print "("; printValue e1; print " /\\ "; printValue e2; print ")")
	   | OR  => (print "("; printValue e1; print " \\/ "; printValue e2; print ")")
	   | EQ  => (print "("; printValue e1; print " = "; printValue e2; print ")")
	   | NE  => (print "- ("; printValue e1; print " = "; printValue e2; print ")")
	   | LT  => (print "("; printValue e1; print " lt "; printValue e2; print ")")
	   | LE  => (print "("; printValue e1; print " le "; printValue e2; print ")")
	   | GT  => (print "("; printValue e1; print " gt "; printValue e2; print ")")
	   | GE  => (print "("; printValue e1; print " ge "; printValue e2; print ")")
	   | ADD => (print "("; printValue e1; print " + "; printValue e2; print ")")
	   | MUL => (print "("; printValue e1; print " * "; printValue e2; print ")"))
      | _ => print "?";  (* should not happen *)

fun printProbe pb =
    case pb of
        StateProbe (_,s) => print s
      | ValueProbe (_,e) => printValue e
      | EventProbe (_,l) => print l
      | TagProbe (_,l)   => print l
      | NotProbe r => (print "("; print "- "; printProbe r; print ")")
      | AndProbe (r1,r2) => (print "("; printProbe r1; print " /\\ "; printProbe r2; print ")")
      | OrProbe (r1,r2) => (print "("; printProbe r1; print " \\/ "; printProbe r2; print ")")
      | ImpProbe (r1,r2) => (print "("; printProbe r1; print " => "; printProbe r2; print ")")
      | _ => print "?";  (* should not happen *)

fun printProp p =
    case p of
        DeadlockFreeProp => print "deadlockFree"
      | InfinitelyOftenProp p => (print "infinitelyOften "; printProbe p)
      | MortalProp p => (print "mortal "; printProbe p)
      | LiveProp p => (print "live "; printProbe p)
      | PresentProp (p,NONE) => (print "present "; printProbe p)
      | PresentAfterProp (p1,p2,NONE) => (print "presentAfter "; printProbe p1; print " "; printProbe p2)
      | PresentBeforeProp (p1,p2,NONE) => (print "presentBefore "; printProbe p1; print " "; printProbe p2)
      | PresentBetweenProp (p1,p2,p3,NONE) => (print "presentBetween "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | PresentUntilProp (p1,p2,p3,NONE) => (print "presentUntil "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | AbsentProp (p,NONE) => (print "absent "; printProbe p)
      | AbsentAfterProp (p1,p2,NONE) => (print "absentAfter "; printProbe p1; print " "; printProbe p2)
      | AbsentBeforeProp (p1,p2,NONE) => (print "absentBefore "; printProbe p1; print " "; printProbe p2)
      | AbsentBetweenProp (p1,p2,p3,NONE) => (print "absentBetween "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | AbsentUntilProp (p1,p2,p3,NONE) => (print "absentUntil "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | LeadstoProp (p1,p2,NONE) => (print "leadsto "; printProbe p1; print " "; printProbe p2)
      | LeadstoBeforeProp (p1,p2,p3,NONE) => (print "leadstoBefore "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | LeadstoAfterProp (p1,p2,p3,NONE) =>  (print "leadstoAfter "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | LeadstoBetweenProp (p1,p2,p3,p4,NONE) => (print "leadstoBetween "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3; print " "; printProbe p4)
      | LeadstoUntilProp (p1,p2,p3,p4,NONE) => (print "leadstoUntil "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3; print " "; printProbe p4)
      | AlwaysProp (p,NONE) => (print "always "; printProbe p)
      | AlwaysAfterProp (p1,p2,NONE) => (print "alwaysAfter "; printProbe p1; print " "; printProbe p2)
      | AlwaysBeforeProp (p1,p2,NONE) => (print "alwaysBefore "; printProbe p1; print " "; printProbe p2)
      | AlwaysBetweenProp (p1,p2,p3,NONE) => (print "alwaysBetween "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | AlwaysUntilProp (p1,p2,p3,NONE) => (print "alwaysUntil "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | PrecedesProp (p1,p2,NONE) => (print "precedes "; printProbe p1; print " "; printProbe p2)
      | PrecedesBeforeProp (p1,p2,p3,NONE) => (print "precedesBefore "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | PrecedesAfterProp (p1,p2,p3,NONE) => (print "precedesAfter "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3)
      | PrecedesBetweenProp (p1,p2,p3,p4,NONE) => (print "precedesBetween "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3; print " "; printProbe p4)
      | PrecedesUntilProp (p1,p2,p3,p4,NONE) => (print "precedesUntil "; printProbe p1; print " "; printProbe p2; print " "; printProbe p3; print " "; printProbe p4)
      | AndProp (p1,p2) => (print "("; printProp p1; print " /\\ "; printProp p2; print ")")
      | OrProp (NotProp p1,p2) => (print "("; printProp p1; print " => "; printProp p2; print ")")
      | OrProp (p1,p2) => (print "("; printProp p1; print " \\/ "; printProp p2; print ")")
      | ImpProp (p1,p2) => (print "("; printProp p1; print " => "; printProp p2; print ")")
      | NotProp p => (print "- "; print "("; printProp p; print ")")
      | BoxProp p => (print "[] "; print "("; printProp p; print ")")
      | DiaProp p => (print "<> "; print "("; printProp p; print ")")
      | NextProp p => (print "() "; print "("; printProp p; print ")")
      | UntilProp (p1,p2) => (print "("; printProp p1; print " U "; printProp p2; print ")")
      | ReleaseProp (p1,p2) => (print "("; printProp p1; print " V "; printProp p2; print ")")
      | ProbeProp p => (print "("; printProbe p; print ")")
      | VarProp n => print n
      | _ => (* timed properties *) print "?timed?";


fun printProperty (PropertyDecl (s,p)) = (newline 0; print "op "; print s; print " = "; printProp p; print ";")
  | printProperty (Assert s) = (newline 0; print s; print ";");

fun emitFiacre (Program (_::[(_,dl,_)])) =
    ((* insert patterns.ltl library into file *)
     print (case !LOGIC of "selt" => ltllib | "muse" => mmclib | _ => failwith "unknown logic");
     case !DEADTAGS of
	 [] => ()
       | l => (newline 0; newline 0; print "# dead transitions possibly referred to in user formulas";
	       List.app (fn g => (newline 0; print "op "; print g; print " = F;")) l);
     case dl of
	 [] => ()
       | _ => (newline 0; newline 0; print "# user formulas";
	       List.app printProperty dl);
     newline 0; newline 0)
    handle BYE => raise BYE | _ => failwith "emit tl formulas";

end; (* structure TtsTlEmit *)





(* toplevel emit routine *)
signature EMIT = sig
val emit : TTS -> unit
end;

structure Emit : EMIT = struct
fun emit t =
    case !OUTPUT of
	TTS => let val dir = !OutPrint.name
		   val base = #base(OS.Path.splitBaseExt (#file (OS.Path.splitDirFile dir)))
		   val fnet = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base=base,ext=SOME "net"}}
		   val snet = TextIO.openOut fnet handle x => emit_error ("cannot open "^fnet)
		   val fdat = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base=base,ext=SOME "c"}}
		   val sdat = TextIO.openOut fdat handle x => emit_error ("cannot open "^fdat)
		   val fdef = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base=base,ext=SOME "h"}}
		   val sdef = TextIO.openOut fdef handle x => emit_error ("cannot open "^fdef)
	       in OutPrint.name := fnet;
		  OutPrint.FILE := snet;
		  printVerbose "::Writing Net(.net) File\n";
		  TtsNetEmit.emitFiacre t;
		  printVerbose "::Done\n";
		  let (* ugly but will improve later *)
		      fun writeh f x =
			  (OutPrint.flush (); OutPrint.name := fdef; OutPrint.FILE := sdef;
			   f x;
			   OutPrint.flush (); OutPrint.name := fdat; OutPrint.FILE := sdat)
		  in OutPrint.name := fdat;
		     OutPrint.FILE := sdat;
		     printVerbose "::Writing Data(.h|.c) Files\n";		 
		     TtsCEmit.emitFiacre (base^".h",fdef,writeh) t;
		     printVerbose "::Done\n"
		  end;
		  if !PROPERTIES then
		      let val ext = case !LOGIC of "selt" => "ltl" | "muse" => "mmc" | _ => "?" (* should not happen *)
			  val fltl = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base=base,ext=SOME ext}}
			  val sltl = TextIO.openOut fltl handle x => (ErrPrint.print ("cannot open "^fltl^"\n"); raise BYE)
		      in OutPrint.name := fltl;
			 OutPrint.FILE := sltl;
			 printVerbose ("::Writing Ltl (." ^ ext ^ ") Files\n");		 
			 TtsLtlEmit.emitFiacre t;
			 printVerbose "::Done\n"
		      end
		  else
		      ()
	       end
      | Abstract => (* for frac -G *)
	        (printVerbose "::Writing Abstract tts\n";
		 OutTtsEmit.emitTts t;
		 printVerbose "::Done\n")
      | _ => ErrPrint.print ("unknown code generator\n");
end;


