
(* output options *)
datatype output = NoOutput | Fiacre | TTS | Abstract | LOTOS;
val OUTPUT = ref NoOutput;  (* output option, default just parses *)

(* mode flags *)
val VERBOSE = ref false;
val DETAILS = ref false;     (* for debugging specializations, with flags -g *)
val MAXIMIZE = ref true;     (* debug option for type instances of primitives *)
val TYPECHECK = ref true;    (* typechecks or not *)
val TRANSF = ref false;      (* if -F, prints result of pre-processing rather than fiacre sources *)
val ARCH = ref false;        (* print instances tree if -arch *)
val KEEPTRANS = ref false;   (* enable/disable dead and dup transition removal *)
val NOPRIOCHECK = ref false; (* enable/disable priority relation check *)
val AADL = ref false;        (* for compat with 1.1.7 *)
val OPTIMIZE = ref true;     (* apply reduce o simplify or not *)
val PROPERTIES = ref true;   (* does or not not handle properties *)
val PROBES = ref false;      (* does or not handle probes *)
val STRIP = ref false;       (* forget property extensions on the fly *)
val LOGIC = ref "selt";      (* logic/tool for properties output *)
val SHARING = ref true;      (* share or not arrays and records state components *)
val UNSAFE = ref false;   (* skips some sanity tests or transforms them into warnings ... *)


(* error handling *)
exception BYE;  (* signals termination *)
fun error m = (ErrPrint.print m; ErrPrint.print "\n"; ErrPrint.flush (); raise BYE);
fun warning m = (ErrPrint.print m; ErrPrint.print "\n"; ErrPrint.flush ());

(* banner *)
val BANNER = "Frac version "^version^release^" -- "^date^" -- FERIA/SVF & INRIA/VASY\n";

(* removes "/" at end of file, if any *)
fun strip f =
    let val k = String.size f
    in if #"/" = String.sub(f,k-1) then String.substring (f,0,k-1) else f
    end;

(* command line parser *)
fun command args =
	let open OutPrint
	    fun usage () =
		(print "usage: frac [-h | -help]\n";
		 print "            [-q | -v] [-o errfile] [-p | -c | -f | -t]\n";
		 print "            [-flat | -share] [-unsafe] [-O | -g] [-strip] [-arch]\n";
		 print "            [infile] [outfile]\n")
	    fun fracerror m =
	        (print "frac: "; print m; print "\n"; usage(); raise BYE)
	    fun help () =
		(print BANNER;
		 usage();
		 print "FLAGS        WHAT                                     DEFAULT\n";
		 print "-h | -help   this mode                                \n";
	         print "error format:                                         \n";
		 print "-q            quiet                                   -q\n";
		 print "-v            verbose                                 -q\n";
		 print "-o errfile    save errors in errfile (stderr if -)    stderr\n";
		 print "output format:                                        \n";
		 print "-p            just parses                             -c\n";
		 print "-c            just parses and checks                  -c\n";
		 print "-f | -fcr     checks then outputs fiacre source       -c\n";
		 print "              (implicit if outfile has extension .fcr)\n";
		 print "-t | -tts      checks then compiles to .tts           -c\n";
		 print "              (implicit if outfile has extension .tts)\n";
 		 print "code generation options:                              \n";
 		 print "-share        enables sharing of state components     -share\n";
 		 print "-flat         disables sharing of state components    -share\n";
 		 print "-unsafe       warns but dont fail on shv interferences\n";
		 print "-O | -g       optimizes or not the code generated     -O\n";
 		 print "other options:                                        \n";
		 print "-strip        forgets properties and assert directives\n";
		 print "-arch         prints component hierarchy              \n";
		 print "files:                                                \n";
		 print "infile        input file (stdin if -)                 stdin\n";
		 print "outfile       output file (stdout if -)               stdout\n";
		 flush ();
		 OS.Process.exit OS.Process.success)
	    val FILES = ref 0
	    fun concat l = String.concat (map (fn x => " "^x) l)
	    fun parse l = case l of
		[] => ()
	      | "-help"::_ => help()
	      | "-h"::_    => help()
	      | "-q"::t    => (VERBOSE := false; parse t)
	      | "-v"::t    => (VERBOSE := true; parse t)
	      | "-o"::file::t => (if String.size file > 1 andalso "-" = String.substring(file,0,1)
				  then fracerror ("bad command line, skipping "^file^" ... "^concat t)
				  else ErrPrint.name := strip file;
				  parse t)
	      | "-p"::t    => (OUTPUT := NoOutput; TYPECHECK := false; parse t)
	      | "-c"::t    => (OUTPUT := NoOutput; TYPECHECK := true; parse t)
	      | "-f"::t    => (OUTPUT := Fiacre; parse t)
	      | "-fcr"::t  => (OUTPUT := Fiacre; parse t)
	      | "-t"::t    => (OUTPUT := TTS; parse t)
	      | "-tts"::t  => (OUTPUT := TTS; parse t)
(* private *) | "-F"::t    => (OUTPUT := Fiacre; TRANSF := true; parse t)
(* private *) | "-Fd"::t   => (OUTPUT := Fiacre; TRANSF := true; DETAILS := true; parse t)
(* private *) | "-G"::t    => (OUTPUT := Abstract; parse t)
(* private *) | "-Gd"::t   => (OUTPUT := Abstract; DETAILS := true; parse t)
(* private *) | "-keeptrans"::t => (KEEPTRANS := true; parse t)
(* private *) | "-nopriocheck"::t => (NOPRIOCHECK := true; parse t)
(* private *) | "-arch"::t => (ARCH := true; parse t)
(* private *) | "-g"::t    => (OPTIMIZE := false; parse t)
(* private *) | "-O"::t    => (OPTIMIZE := true; parse t)
	      | "-strip"::t => (STRIP := true; parse t)
	      | "-obs"::t => (PROBES := true; parse t)
	      | "-properties"::t => (PROPERTIES := true; parse t)
	      | "-selt"::t => (LOGIC := "selt"; parse t)
	      | "-muse"::t => (LOGIC := "muse"; parse t)
(* obsolete *)| "-aadl"::t   => (AADL := true; parse t)
	      | "-unsafe"::t => (UNSAFE := true; parse t)	
 	      | "-share"::t => (SHARING := true; parse t)
 	      | "-flat"::t => (SHARING := false; parse t)
	      | file::t    => if String.size file > 1 andalso "-" = String.substring(file,0,1)
			      then fracerror ("bad command line, skipping "^file^" ... "^concat t)
	                      else (FILES := 1 + !FILES; case !FILES of
			    1 => if file = "-" then parse t else (Input.name := strip file; parse t)
			  | 2 => if file = "-" then parse t else (OutPrint.name := strip file; parse t)
			  | _ => fracerror ("bad command line, skipping "^file^" ... "^concat t))
	in parse args
	end;




