
(* removes ports from code, expands probes if any, then normalizes guards *)


signature POSTCOMPOSE = sig
    val postcompose : TTS.TTS -> TTS.TTS
end;


structure PostCompose : POSTCOMPOSE = struct

(* for probes *)

(* for numbering probe vars *)
val CNT = ref 0;

(* for computing priority relation between sections *)
val SECTIONS = ref [];

(* declares if needed and returns aux variable associated with an enter/leave probe *)
fun enterleaveVar root f initial =
    let fun newVars () = (CNT := 1 + !CNT; (root ^ "_p" ^ Int.toString (!CNT),root ^ "_c" ^ Int.toString (!CNT)))
    in case List.find (fn (_,_,_,_,g) => g=f) (!ENTERLEAVEVARS) of
	   SOME (_,_,v,_,_) => v
	 | NONE => let val (tmpv,v) = newVars ()
		   in ENTERLEAVEVARS := (root,tmpv,v,initial,f) :: !ENTERLEAVEVARS;
		      v
		   end
    end;

(* declares if needed and returns aux variable associated with an enter/leave probe *)
fun valueVar root f =
    let fun newVar () = (CNT := 1 + !CNT; (root ^ "_w" ^ Int.toString (!CNT)))
    in case List.find (fn (_,_,g) => g=f) (!VALUEVARS) of
	   SOME (_,v,_) => v
	 | NONE => let val v = newVar ()
		   in VALUEVARS := (root,v,f) :: !VALUEVARS;
		      v
		   end
    end;

(* computes an enter/leave probe on auxiliary variables *)
fun probeVarCond pb =
    case pb of
	StateProbe ((root,_),s) => comperror "unexpected state probe in probeVarCond" (* converted to ValueProbe in compose *)
      | ValueProbe ((root,_),p) => VarExp(enterleaveVar root p (BoolExp false),BoolType)
      | EventProbe ((root,_),_) => comperror "event probes under enter|leave forbidden"
      | TagProbe ((root,_),_)   => comperror "tag probes under enter|leave forbidden"
      | NotProbe pb => UnopExp (NOT,BoolType,probeVarCond pb)
      | AndProbe (pb1,pb2) => InfixExp (AND,BoolType,probeVarCond pb1,probeVarCond pb2)
      | OrProbe (pb1,pb2) => InfixExp (OR,BoolType,probeVarCond pb1,probeVarCond pb2)
      | ImpProbe (pb1,pb2) => InfixExp (OR,BoolType,UnopExp(NOT,BoolType,probeVarCond pb1),probeVarCond pb2)
      | EnterProbe pb => comperror "embedded enter|leave probes forbidden"
      | LeaveProbe pb => comperror "embedded enter|leave probes forbidden";



(* remembers if some transition is timed *)
val TIMED = ref false;

(* attaches time constraints to tts transitions *)
(* for each transition: get the port used in the transition, if any, then store its time interval within the transition *)
(* also maintains TIMED *)
fun insertTime (ProcessDecl (b,c,d,e,_,k,f,h,i,init,trl)) =
    (let fun getPortName s1 =
	     case s1 of
		 StSequence(a1,a2) => (case getPortName a1 of  "" => getPortName a2 | st => st)
	       | StSignal pn => pn
	       | StInput(pn,_,_) => pn
	       | StInputList(pn,_) => pn
	       | StOutput(pn,_,_) => pn
	       | _ => ""
	 fun updateBound (tag,st,stm,bnd,prl) = 
             let val port = getPortName stm
	     in TIMED := (!TIMED orelse nontrivial bnd);
		case List.find (fn (x,_,_,_) => x=port) h of
		    SOME (_,_,_,tc) => (TIMED := (!TIMED orelse nontrivial tc); (tag,st,stm,tc,prl))
		  | NONE => (tag,st,stm,bnd,prl)
             end
     in ProcessDecl (b,c,d,e,[],k,f,
		     List.filter (fn (_,_,SOME _,_) => true | _ => false) h, (* probes kept *)
		     i,init,
		     map updateBound trl)
     end
     handle BYE => raise BYE | _ => failwith "insertTime")
  | insertTime d = d;



(* names transitions and replace labels in prio relations by transition names *)
(* also adds section name to user tags of all transitions for later addition of inter-section priority constraints *)
fun setPriorities sec (d as ProcessDecl (name,ports,args,states,rtemp,rwtemp,stvars,locports,prio,init,trl)) =
    (let (* maps elementary tags to transition names *)
	 (* also adds sec to user tags *)
	 (* should use a tree rather than an a-list ? *)
	 val dict = ref []
	 fun record tg name = case List.find (fn (g,_) => g=tg) (!dict) of SOME (_,r) => r := name :: !r | _ => dict := (tg,ref [name]) :: !dict
	 fun get tg = let val ref l = assoc tg (!dict) in l end
	 (* first step: build transition names and record them in dict, also adds sec to atags *)
	 val trl' = map (fn ((tags,lab,atg),fr,stm,bnd,prl) =>
			    let val trname = String.concatWith "_" tags
			    in List.app (fn tg => record tg trname) tags;
			       (([trname],lab,sec :: atg),fr,stm,bnd,prl)
			    end) trl
	 (* second step: replace labels in priority lists by transition names *)
	 val trl'' = map (fn (namelab,fr,stm,bnd,prl) => (namelab,fr,stm,bnd,makeset (flat (map get prl)))) trl'
     in ProcessDecl (name,ports,args,states,rtemp,rwtemp,stvars,locports,prio,init,trl'')
     end handle BYE => raise BYE | _ => failwith "setPriorities")
  | setPriorities _ d = d;



(* sets inter-section priorities. Since sections may be synchronized, relies on section tags added by setPriorities *)
fun setInterPriorities (s as (n,dl,ProcessDecl (name,ports,args,states,rtemp,rwtemp,stvars,locports,prio,init,trl))) =
    (case !INTERPRIO of
	[] => s
      | ipr =>
	let (* secname -> transition names *)
	    val sets = List.map (fn y => (y,List.mapPartial (fn (([trname],_,tags),_,_,_,_) =>
							     if mem y tags then SOME trname else NONE) trl))
				(makeset (map snd (!INTERPRIO)))
	    fun insertPrios [] trl = trl
	      | insertPrios ((x,y)::t) trl =
		insertPrios t (List.map (fn tr as ((tg,lb,tags),fr,stm,bnd,prl) =>
					   if mem x tags then ((tg,lb,tags),fr,stm,bnd,makeset (prl @ assoc y sets)) else tr) trl)
	in (n,dl,ProcessDecl (name,ports,args,states,rtemp,rwtemp,stvars,locports,prio,init,insertPrios (!INTERPRIO) trl))
	end
	handle BYE => raise BYE | _ => failwith "setInterPriorities")
  | setInterPriorities s = s;



(* replaces communication statements by null;
   inserts code for probes;
   concats tags into a single name for the transition;
   builds priority relation, replacing elem tags by transition names
*)
fun replacePorts (ProcessDecl (b,c,d,e,_,k,f,h,i,init,trl)) =
    (let (* number of probe vars on entry (needed to avoid duplicated declarations) *)
	(* unify patterns, returns nuber of solutions (0, 1, or 2 meaning "more than 1") *)
	fun unify_patterns stll =
	    let fun isConst p =
		    case p of
			BoolExp _ => true
		      | IntExp _ => true
		      | EnumExp _ => true
		      | ConExp _ => true
		      | ConstrExp (_,p,_) => isConst p
		      | _ => false
		exception NO
		fun unify p1 p2 =
		    case (p1,p2) of
			(BoolExp b1,BoolExp b2) => if b1 = b2 then p1 else raise NO
		      | (IntExp (i1,_),IntExp (i2,_)) => if i1 = i2 then p1 else raise NO
		      | (EnumExp (c1,_),EnumExp (c2,_)) => if c1 = c2 then p1 else raise NO
		      | (ConExp (c1,_),ConExp (c2,_)) => if c1 = c2 then p1 else raise NO
		      | (ConstrExp (c1,p1,ty), ConstrExp(c2,p2,_)) => if c1 = c2 then ConstrExp (c1,unify p1 p2,ty) else raise NO
		      | (VarExp _,_) => p2
		      | (_,VarExp _) => p1
		      | (AnyExp l1, AnyExp l2) => AnyExp (l1@l2)
		      | (AnyExp _,_) => p2
		      | (_,AnyExp _) => p1
		      | _ => AnyExp []
		fun list_unify [h] = h
		  | list_unify (h::t) = map (fn (p1,p2) => unify p1 p2) (ListPair.zip (h,list_unify t))
		val pl = list_unify stll
	    in (if List.all isConst pl then 1 else 2,pl)
	    end
	    handle NO => (0,[])
	(*  replaces communication statements by null; inserts code for probes*)
	fun transIO (tglab,fr,stm,bnd,prl) =
            (* root info is for computing the set of roots that must maintain probe variables *)
	    let (* prove value *)
		fun probeCond pb =
		    case pb of
			StateProbe ((root,_),s) => comperror "unexpected state probe in probeCond" (* converted to ValueProbe in compose *)
		      | ValueProbe ((root,_),p) => p
		      | EventProbe ((root,_),_) => comperror "event probes as ports unsupported yet, sorry"
		      | TagProbe ((root,_),_)   => comperror "tag probes as ports unsupported yet, sorry"
		      | NotProbe pb => UnopExp (NOT,BoolType,probeCond pb)
		      | AndProbe (pb1,pb2) => InfixExp (AND,BoolType,probeCond pb1,probeCond pb2)
		      | OrProbe (pb1,pb2) => InfixExp (OR,BoolType,probeCond pb1,probeCond pb2)
		      | ImpProbe (pb1,pb2) => InfixExp (OR,BoolType,UnopExp(NOT,BoolType,probeCond pb1),probeCond pb2)
		      | EnterProbe pb => InfixExp (AND,BoolType,UnopExp (NOT,BoolType,probeVarCond pb),probeCond pb)
		      | LeaveProbe pb => InfixExp (AND,BoolType,probeVarCond pb,UnopExp (NOT,BoolType,probeCond pb))
		fun replaceIO s =
		    case s of
			StSignal p =>
			(case List.find (fn (q,_,SOME _,_) => q=p | _ => false) h of
			     SOME (_,_,SOME pb,_) => StGuard (probeCond pb)
			   | NONE => StNull)
		      | StOutput (p,_,expl) =>
			if List.exists (fn e => case e of AnyExp (_::_) => true | _ => false) expl then
			    comperror ("unsupported open input on port "^p)
			else
			    StNull
		      | StInput (p,_,stl) => replaceIO (StInputList (p,[stl]))
		      | StInputList (p,stll) =>
			(* computes unification of all patterns; reject if more than one solution *)
			(* Note: works only if all conditions captured by patterns, not "where" clauses *)
			let val (k,pl1) = unify_patterns stll
			in case k of
			       0 => StGuard (BoolExp false)  (* handled; will be removed *)
			     | 1 => (* handled; will be kept or removed *)
			       List.foldr StSequence StNull (flat (map (fn stList => map (fn (x,y) => ToTTS.convAssignMatch(StAssign(x,y)))
											 (ListPair.zip (stList,pl1)))
								       stll))
			     | 2 => (* unsupported yet *) comperror ("unsupported open input on port "^p)
			end
		      | _ => s
		val stm' = mapStm replaceIO stm
	    in (tglab,fr,stm',bnd,prl)
	    end
    in ProcessDecl (b,c,d,e,[],k,f,h,i,init,map transIO trl)
    end
    handle BYE => raise BYE | _ => failwith "replacePorts")
  | replacePorts d = d;
                                                                                          


(* moves conditional statements after assignements to front. *)
(* general form (S is state):
      S := f S; if c S then st end ==> if c (f S) then S := f S; st end
   example 1: (independant)
      Z := 3; if X>Y then s end ==> if X>Y then Z := 3; s end
   example 2: (not independant), with analysis: (not implemented)
      a[z]:=3; if a[x]>a[y] then s end ==> if ((x=z) ? 3 : a[x]) > ((y=z) ? 3 : a[y]) then a[z] := 3; s end
   example 2.2: (not independant), with state copying:
      a[z]:=3; if a=b then s end ==> if (a[z]:=3;a=b) then s end
   example 3: (not independant but substitutable)
      X := Y+3; if X>Y then s end ==> if Y+3>Y then X := Y+3; s end
*)
fun normGuards (ProcessDecl (b,c,d,e,_,k,f,h,i,init,g)) =
    let fun swap s =
	    let fun swapGuard s =
		    case s of
			StSequence (StGuard (BoolExp true),s2) => swapGuard s2
		      | StSequence (StGuard c1,s2) =>
			(case swapGuard s2 of
			     StSequence(StGuard (BoolExp true), s2') => StSequence(StGuard c1, s2')
			   | StSequence(StGuard c2, s2') => StSequence (StGuard (InfixExp(AND,BoolType,c1,c2)), s2'))
		      | StSequence (StNull,s2) => swapGuard s2
		      | StSequence (s1 as StAssign(p,e),s2) =>
			if equal p e then swapGuard s2  else
			let val StSequence(StGuard c2,s2') = swapGuard s2
			    val ws1 = writtenInSt s1
			    val rc = readInExp c2
			    val rs1 = readInSt s1
			    val wc = writtenInExp c2
			in if List.all (fn v => not (mem v wc)) rs1 andalso
			      List.all (fn v => not (mem v ws1)) rc andalso
			      List.all (fn v => not (mem v wc)) ws1
			   then (* independant  *) StSequence (StGuard c2, StSequence(s1,s2'))
			   else (* *)
			        if (* (R(s1) union W(s1)) inter W(c2) empty *)
				   List.all (fn v => not (mem v wc)) (rs1 @ ws1) andalso
				   not (matchExp (fn LetExp _ => true | _ => false) c2) andalso
				   (* p is a variable, not an access ! *)
				   (case p of VarExp _ => true | _ => false) andalso    
				   (* no function in e [may be costly otherwise] *)
				   not (matchExp (fn AppExp _ => true | CallExp _ => true | _ => false) e)
				then (* substitutable *)
				     StSequence (StGuard (mapExp (fn x => if equal p x then e else x) c2), StSequence(s1,s2'))
				else (* *)
				     (* general case  *)
				     StSequence (StGuard (LetExp(s1,c2)), StSequence(s1,s2'))
			end
		      | StSequence (s1,s2) => (* s1 is foreach, while, if *)
			let val StSequence(StGuard c2,s2') = swapGuard s2
			    val ws1 = writtenInSt s1
			    val rc = readInExp c2
			    val rs1 = readInSt s1
			    val wc = writtenInExp c2
			in if List.all (fn v => not (mem v wc)) rs1 andalso
			      List.all (fn v => not (mem v ws1)) rc andalso
			      List.all (fn v => not (mem v wc)) ws1
			   then (* independant  *) StSequence (StGuard c2, StSequence(s1,s2'))
			   else (* general case *) StSequence (StGuard (LetExp(s1,c2)), StSequence(s1,s2'))
			end
		      | _ => StSequence (StGuard (BoolExp true),s)
	    in swapGuard (#1 (ToTTS.normSequence s))  (* normSequence needed for e.g. abp.fcr, due to replacePorts introducing StNull ? *)
	    end
    in ProcessDecl (b,c,d,e,[],k,f,h,i,init,map (fn (tg,fr,s,b,p) => (tg,fr,swap s,b,p)) g)
       handle BYE => raise BYE | _ => failwith "normGuards"
    end
 | normGuards d = d;



(* adds code to transitions of observed processes for implementing enter/leave/value probes *)
(* must be called AFTER replacePorts applied on all sections and translateProperties (so all auxiliary vars are known) *)
fun instrument sect (p as ProcessDecl (name,ports,args,states,rtemp,rwtemp,stvars,locports,prio,init,trl)) =
    if !PROPERTIES orelse !PROBES then
	let fun mkSt [] = StNull | mkSt [s] = s | mkSt (h::t) = StSequence (h,mkSt t)
	    (* concats two statements preserving normal form for sequences (left not a sequence) *)
	    fun join (s1,StNull) = s1
	      | join (StNull,s2) = s2
	      | join (s1 as StTo _,s2) = join (s2,s1)  (* beware: there may be StTo left after compose !! *)
	      | join (StSequence (s11,s12),s2) = join (s11,join (s12,s2))
	      | join (s1,s2) = StSequence (s1,s2)
	    val trl' =
		(* adds enter/leave probe var code to transitions.
		   handling section sect:
		   for each enter/leave var (root,tmpv,v,initval,value) such that root = sect:
	              for each transition:
		         add at beginning of transition:
			     tmpv := value
		         add at end of transition:
		             v := tmpv
	        *) 
		case List.filter (fn (root,_,_,_,_) => root=sect) (!ENTERLEAVEVARS) of
		    [] => trl
		  | pbvars => let val bef = mkSt (List.map (fn (_,tmpv,_,_,f) => StAssign (VarExp(tmpv,BoolType),f)) pbvars)
				  val aft = mkSt (List.map (fn (_,tmpv,v,_,_) => StAssign (VarExp(v,BoolType),VarExp(tmpv,BoolType))) pbvars)
			      in map (fn (tg,fr,s,b,p) => (tg,fr,join(bef,join(#1 (ToTTS.normSequence s),aft)),b,p)) trl
			      end
	    val trl'' = 
		(* adds value probe var code to transitions (only for value probes occurring in properties).
		   handling section sect:
		   for each valuevar (root,v,value) such that root = sect:
	              for each transition:
		         add at end of transition:
		             v := value
		 *)
		case List.filter (fn (root,_,_) => root=sect) (!VALUEVARS) of
		    [] => trl'
		  | pbvars => let val aft = mkSt (List.map (fn (_,v,e) => StAssign (VarExp(v,BoolType),e)) pbvars)
			      in map (fn (tg,fr,s,b,p) => (tg,fr,join(#1 (ToTTS.normSequence s),aft),b,p)) trl'
			      end
	in ProcessDecl (name,ports,args,states,rtemp,rwtemp,stvars,locports,prio,init,trl'')
	end
	handle BYE => raise BYE | _ => failwith "instrument"
    else
	p
  | instrument _ p = p;


(* adds code to transitions of observed processes for implementing enter/leave probes and value probes in properties *)
(* for untimed properties:
     adds auxiliary and temporary variables for each enter/leave probe
     adds an auxiliary variable for each value probe that cannot be translated into a selt property
   then instruments transitions of section sect to maintain these variables
*)
fun translateProperties (p as PropertyDecl (name,prop)) =
    if !PROPERTIES then
	let (* returns e if value probe with contents e can be encoded in selt, or define and auxiliary variable otherwise *)
	    (* THIS SHOULD BE IMPROVED:
	       if e can be translated into selt, then the fiacre variables occurring in the value probe
	       will always be considered state variables (kept in states), while they may be constant or temporaries,
	       this because property bodies are not considered in the opmization pass.
               So there is no guarantee that state size is smaller when an auxiliary variables is avoided ...
	       Ideally, properties should be considered in optimization pass (SEE LATER).
	     *)
	    fun seltencode root e =
		let exception SeltEncode
		    val kept = ref [] (* fiacre variables used in property *)
		    fun intencode root e =
			(* note: preserve types, so instrumentation code may fail ... *)
			case e of
			    InfixExp (f,ty,e1,e2) => if mem f [ADD,MUL] then
							 InfixExp (f,ty,intencode root e1,intencode root e2)
						     else
							 raise SeltEncode
			  | UnopExp (PLUS,ty,ee) => intencode root ee
			  | UnopExp (MINUS,ty,ee) => UnopExp(MINUS,ty,intencode root ee)
			  | VarExp (v,ty) => (case ty of IntervalType _ => (kept := v :: !kept; e)
						       | NatType => (kept := v :: !kept; e)
						       | IntType => (kept := v :: !kept; e)
						       | _ => raise SeltEncode)
			  | IntExp (i,_) => e
			  | _ => raise SeltEncode
		    fun boolencode root e =
			case e of
			    InfixExp (f,ty,e1,e2) => if mem f [AND,OR] then (* bool operands *)
							 InfixExp (f,ty,boolencode root e1,boolencode root e2)
						     else if mem f [EQ,NE,LT,LE,GT,GE] then (* int operands *)
							 InfixExp (f,ty,intencode root e1,intencode root e2)
						     else
							 raise SeltEncode
			  | UnopExp (NOT,ty,ee) => UnopExp (NOT,ty,boolencode root ee)
			  | VarExp (v,BoolType) => (kept := v :: !kept; e)
			  | BoolExp b => e
			  | _ => raise SeltEncode
		    val p = boolencode root e handle SeltEncode => (kept := []; VarExp(valueVar root e,BoolType))
		in List.map (fn v => if mem v (!KEPTVARS) then () else KEPTVARS := v :: !KEPTVARS) (!kept);
		   p
		end
	    (* returns property associated with an observable *)
	    fun probeProp pb =
		case pb of
		    StateProbe _ => pb
		  | ValueProbe ((root,_),p) => ValueProbe (("",[]),seltencode root p)
		  | EventProbe ((root,_),l) => (* \/ of all transitions in root with that label *)
		    let val SOME (_,ProcessDecl (_,_,_,_,_,_,_,_,_,_,trl)) = List.find (fn (sec,_) => sec=root) (!SECTIONS)
			fun mkOr [x] = x | mkOr (h::t) = OrProbe (h,mkOr t)
		    in mkOr (List.mapPartial (fn (([tag],lab,_),_,_,_,_) => if lab=l then SOME (EventProbe (("",[]),tag)) else NONE) trl)
		    end
		  | TagProbe ((root,_),g) => (* \/ of all transitions in root with that tag *)
		    let val SOME (_,ProcessDecl (_,_,_,_,_,_,_,_,_,_,trl)) = List.find (fn (sec,_) => sec=root) (!SECTIONS)
			fun mkOr [x] = x | mkOr (h::t) = OrProbe (h,mkOr t)
		    in mkOr (List.mapPartial (fn (([tag],_,atg),_,_,_,_) => if mem g atg then SOME (TagProbe (("",[]),tag)) else NONE) trl)
		    end
		  | NotProbe pb => NotProbe (probeProp pb)
		  | AndProbe (pb1,pb2) => AndProbe (probeProp pb1,probeProp pb2)
		  | OrProbe (pb1,pb2) => OrProbe (probeProp pb1,probeProp pb2)
		  | ImpProbe (pb1,pb2) => ImpProbe (probeProp pb1,probeProp pb2)
		  | EnterProbe pb => AndProbe (NotProbe (ValueProbe(("",[]),probeVarCond pb)),probeProp pb) (* same as for probes as ports *)
		  | LeaveProbe pb => AndProbe (ValueProbe (("",[]),probeVarCond pb),NotProbe(probeProp pb)) (* same as for probes as ports *)
	    fun propTrans p =
		case p of
		    DeadlockFreeProp => DeadlockFreeProp
		  | InfinitelyOftenProp pb => InfinitelyOftenProp (probeProp pb)
		  | LiveProp pb => LiveProp (probeProp pb)
		  | MortalProp pb => MortalProp (probeProp pb)
		  | PresentProp (pb,NONE) => PresentProp (probeProp pb,NONE)
		  | PresentAfterProp (pb1,pb2,NONE) => PresentAfterProp (probeProp pb1,probeProp pb2,NONE)
		  | PresentBeforeProp (pb1,pb2,NONE) => PresentBeforeProp (probeProp pb1,probeProp pb2,NONE)
		  | PresentBetweenProp (pb1,pb2,pb3,NONE) => PresentBetweenProp (probeProp pb1,probeProp pb2,probeProp pb3,NONE) 
		  | PresentUntilProp (pb1,pb2,pb3,NONE) => PresentUntilProp (probeProp pb1,probeProp pb2,probeProp pb3,NONE)
		  | AbsentProp (pb,NONE) => AbsentProp (probeProp pb,NONE)
		  | AbsentAfterProp (pb1,pb2,NONE) => AbsentAfterProp (probeProp pb1,probeProp pb2,NONE)
		  | AbsentBeforeProp (pb1,pb2,NONE) => AbsentBeforeProp (probeProp pb1,probeProp pb2,NONE)
		  | AbsentBetweenProp (pb1,pb2,pb3,NONE) => AbsentBetweenProp (probeProp pb1,probeProp pb2,probeProp pb3,NONE)
		  | AbsentUntilProp (pb1,pb2,pb3,NONE) => AbsentUntilProp (probeProp pb1,probeProp pb2,probeProp pb3,NONE)
		  | LeadstoProp (pb1,pb2,NONE) => LeadstoProp (probeProp pb1,probeProp pb2,NONE)
		  | LeadstoBeforeProp (pb1,pb2,pb3,NONE) => LeadstoBeforeProp (probeProp pb1,probeProp pb2,probeProp pb3,NONE)
		  | LeadstoAfterProp (pb1,pb2,pb3,NONE) => LeadstoAfterProp (probeProp pb1,probeProp pb2,probeProp pb3,NONE)
		  | LeadstoBetweenProp (pb1,pb2,pb3,pb4,NONE) => LeadstoBetweenProp (probeProp pb1,probeProp pb2,probeProp pb3,probeProp pb4,NONE)
		  | LeadstoUntilProp (pb1,pb2,pb3,pb4,NONE) => LeadstoUntilProp (probeProp pb1,probeProp pb2,probeProp pb3,probeProp pb4,NONE)
		  | AlwaysProp (pb,NONE) => AlwaysProp (probeProp pb,NONE)
		  | AlwaysAfterProp (pb1,pb2,NONE) => AlwaysAfterProp (probeProp pb1,probeProp pb2,NONE)
		  | AlwaysBeforeProp (pb1,pb2,NONE) => AlwaysBeforeProp (probeProp pb1,probeProp pb2,NONE)
		  | AlwaysBetweenProp (pb1,pb2,pb3,NONE) => AlwaysBetweenProp (probeProp pb1,probeProp pb2,probeProp pb3,NONE)
		  | AlwaysUntilProp (pb1,pb2,pb3,NONE) => AlwaysUntilProp (probeProp pb1,probeProp pb2,probeProp pb3,NONE)
		  | PrecedesProp (pb1,pb2,NONE) => PrecedesProp (probeProp pb1,probeProp pb2,NONE)
		  | PrecedesBeforeProp (pb1,pb2,pb3,NONE) => PrecedesBeforeProp (probeProp pb1,probeProp pb2,probeProp pb3,NONE)
		  | PrecedesAfterProp (pb1,pb2,pb3,NONE) => PrecedesAfterProp (probeProp pb1,probeProp pb2,probeProp pb3,NONE)
		  | PrecedesBetweenProp (pb1,pb2,pb3,pb4,NONE) => PrecedesBetweenProp (probeProp pb1,probeProp pb2,probeProp pb3,probeProp pb4,NONE)
		  | PrecedesUntilProp (pb1,pb2,pb3,pb4,NONE) => PrecedesUntilProp (probeProp pb1,probeProp pb2,probeProp pb3,probeProp pb4,NONE)
		  | AndProp (p1,p2) => AndProp (propTrans p1,propTrans p2)
		  | OrProp (p1,p2) => OrProp (propTrans p1,propTrans p2)
		  | ImpProp (p1,p2) => ImpProp (propTrans p1,propTrans p2)
		  | NotProp p => NotProp (propTrans p)
		  | BoxProp p => BoxProp (propTrans p)
		  | DiaProp p => DiaProp (propTrans p)
		  | NextProp p => NextProp (propTrans p)
		  | UntilProp (p1,p2) => UntilProp (propTrans p1,propTrans p2)
		  | ReleaseProp (p1,p2) => ReleaseProp (propTrans p1,propTrans p2)
		  | ProbeProp p => ProbeProp (probeProp p)
		  | VarProp s => VarProp s
		  (* A FAIRE pour propriétés acceptées par le parser: inserer observateurs et traduire la propriété : *)
		  | _                    => comperror ("unsupported property yet (in "^name^"), sorry")
	in PropertyDecl (name,propTrans prop)
	end
    else
	p
  | translateProperties p = p;



(* links the different sections of a fiacre program, adding probes if used *)
(* done before optimization as some variables used later in probes could be removed otherwise *)
fun assemble sl =
    let (* doubles any "_" in name to prevent ambiguity of instantiated names: *)
	fun encode id = String.translate (fn #"_" => "__" | c => Char.toString c) id;
	fun asmSection (ProcessDecl (name1,ports1,args1,states1,rtemp1,rwtemp1,stvars1,locports1,prio1,StTo init1,trl1),
			ProcessDecl (name2,ports2,args2,states2,rtemp2,rwtemp2,stvars2,locports2,prio2,StTo init2,trl2)) =
	    ProcessDecl (encode name1 ^ "_" ^ encode name2,
			 ports1 @ ports2,
			 args1 @ args2,
			 states1 @ states2,
			 rtemp1 @ rtemp2,
			 rwtemp1 @ rwtemp2,
			 stvars1 @ stvars2,
			 locports1 @ locports2,
			 prio1 @ prio2,
			 StTo (init1 @ init2),
			 trl1 @ trl2)
	val (dl',pl as p :: t) = ListPair.unzip (map (fn (s,x,y) => (x,y)) sl)
	val ProcessDecl (name,ports,args,states,rtemp,rwtemp,stvars,locports,prio,init,trl) = List.foldl asmSection p t
    in (* ENTERLEAVEVARS added to process variables, optimizer will find out temporaries *)
	("__main__",
	 flat dl',
	 ProcessDecl (name,ports,args,states,rtemp,rwtemp,
		      stvars @
		      flat (List.map (fn (_,tmpv,v,init,f) => [(tmpv,BoolType,IntExp(0,IntType)),(v,BoolType,init)]) (!ENTERLEAVEVARS)) @
		      flat (List.map (fn (_,v,_) => [(v,BoolType,IntExp(0,IntType))]) (!VALUEVARS)),
		      locports,prio,init,trl))
    end
    handle BYE => raise BYE | _ => failwith "assemble";


(* ##### NOTE: probablement a reordonner si les propriétés introduisent `a la volée des observateurs ##### *)
fun postcompose (Program sl0) =
    let val (sl,slp) = if !PROPERTIES then List.partition (fn ("#properties",_,_) => false | _ => true) sl0 else (sl0,[])
	(* records processed sections in SECTIONS, for computation of inter-section priorities *)
	val sl' = map (fn (s,dl,d) => let val d' = (replacePorts o setPriorities s o insertTime) d
						    in SECTIONS := (s,d') :: !SECTIONS;
						       (s, dl, d')
						    end) sl
	val slp' = case slp of [] => [] | [(s,dl,d)] => [(s,map translateProperties dl,d)]
	val sl'' = map (fn (s,dl,d) => (s, dl, (normGuards o instrument s) d)) sl'
	val p = setInterPriorities (assemble sl'')
    in Program (if !PROPERTIES then p::slp' else [p])
    end
    handle BYE => raise BYE | _ => failwith "postcompose";

end; (* PostCompose *)

